/*------------------------------------------------------------------------------
 bin/djv_info/djv_info.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_base/sys.h>
#include <djv_base/file.h>
#include <djv_base/dir.h>
#include <djv_image/time.h>
#include <djv_image/io.h>
#include <djv_image/app.h>

namespace
{
using namespace djv;

class App : public image::App
{
public:

  App(int, char **);

private:

  void file(const File &, bool path = false) throw (Error);
  void dir(const File &, bool label) throw (Error);

  Str label_time(int, const image::time::Speed &) const;

  void cmdln() throw (Error);
  Str help() const;

  List<Str> _input;
  bool _verbose, _seq, _recurse, _timecode;
  int _columns;
  Str _load;
};

App::App(int argc, char ** argv) :
  image::App("djv_info", argc, argv),
  _verbose(false), _seq(true), _recurse(false), _timecode(true),
  _columns(sys::terminal_width())
{
  DJV_DBG_OBJ("App::App");
    
  // Command line.

  if (app::EXIT_DEFAULT == exit())
    try { cmdln(); }
    catch (Error in)
    {
      error(str::Format("Command Line: %1").arg(in));
      exit(app::EXIT_ERROR);
      return;
    }
  if (exit() != app::EXIT_DEFAULT) return;

  // Setup.

  if (! _input.size()) _input += File(".");
  DJV_DBG_PRN("input = " << _input);
  
  List<File> list;
  for (uint i = 0; i < _input.size(); ++i)
  {
    File file(_input[i]);
    if (_seq && file.seq_valid())
    {
      file.type = file::SEQ;
      for (int j = file.seq.start; j <= file.seq.end; ++j)
        list += file.get(j);
    }
    else
    {
      list += file;
    }
  }
  if (_seq) file::seq_compress(&list);
  DJV_DBG_PRN("files = " << list);
  
  // Work.
  
  for (uint i = 0; i < list.size(); ++i)
    try
    {
      if (file::DIR == list[i].type)
        dir(list[i], (list.size() > 1) || _recurse);
      else
        file(list[i]);
    }
    catch (Error in)
    {
      error(in);
      exit(app::EXIT_ERROR);
    }
}

void App::file(const File & in, bool path) throw (Error)
{
  DJV_DBG_OBJ("App::file");
  DJV_DBG_PRN("in = " << in);
  
  // Open.

  List<image::Info> info;
  image::Tag tag;
  image::Time time;
  try
  {
    std::auto_ptr<image::io::Load>(image::io::load(in, &info, &tag, &time,
      _load));
  }
  catch (Error error)
  {
    throw Error(str::Format("%1: %2").arg(error).arg(in.name()));
  }
  
  // Print.

  Str name = in.get(-1, path);
  
  if (! _verbose)
  {
    const Str str = str::Format("%1x%2:%3 %4 %5@%6").
      arg(info[0].size.x).arg(info[0].size.y).arg(aspect(info[0].size), 0, -2).
      arg(image::pixel::label_pixel()[info[0].pixel]).
      arg(label_time(time.end - time.start + 1, time.speed)).
      arg(image::time::speed_to_float(time.speed), 0, -2);
    print(str::Format("%1 %2").arg(name).
      arg(str, _columns - int(name.size()) - 2));
    
    for (uint i = 1; i < info.size(); ++i)
    {
      const Str name_str = str::Format("#%1 %2").arg(i).arg(info[i].name);
      const Str info_str = str::Format("%1x%2:%3 %4").
        arg(info[i].size.x).arg(info[i].size.y).
        arg(aspect(info[i].size), 0, -2).
        arg(image::pixel::label_pixel()[info[i].pixel]);
      print(str::Format("%1 %2").arg(name_str).
        arg(info_str, _columns - int(name_str.size()) - 2));
    }
  }
  else
  {
    print(name);

    print(str::Format("  Width = %1").arg(info[0].size.x));
    print(str::Format("  Height = %1").arg(info[0].size.y));
    print(str::Format("  Aspect = %1").arg(aspect(info[0].size), 0, -2));
    print(str::Format("  Pixel = %1").
      arg(image::pixel::label_pixel()[info[0].pixel]));
    
    for (uint i = 1; i < info.size(); ++i)
    {
      print(str::Format("  Layer = %1").arg(info[i].name));
      print(str::Format("    Width = %1").arg(info[i].size.x));
      print(str::Format("    Height = %1").arg(info[i].size.y));
      print(str::Format("    Aspect = %1").arg(aspect(info[i].size), 0, -2));
      print(str::Format("    Pixel = %1").
        arg(image::pixel::label_pixel()[info[i].pixel]));
    }

    print(str::Format("  Start = %1").
      arg(label_time(time.start, time.speed)));
    print(str::Format("  End = %1").
      arg(label_time(time.end, time.speed)));
    print(str::Format("  Duration = %1").
      arg(label_time(time.end - time.start + 1, time.speed)));
    print(str::Format("  Speed = %1").
      arg(image::time::speed_to_float(time.speed)));

    const List<Str> key_list = tag.key_list();
    for (uint i = 0; i < key_list.size(); ++i)
      print(str::Format("  Tag %1 = %2").arg(key_list[i]).
        arg(tag[key_list[i]]));

    separator();
  }
}

void App::dir(const File & in, bool label) throw (Error)
{
  DJV_DBG_OBJ("App::dir");
  DJV_DBG_PRN("in = " << in);
  
  // Setup.

  Dir dir(in);
  if (! dir.valid())
    throw Error(str::Format("%1: Open Error").arg(in));

  List<File> list = dir.list();
  if (_seq) file::seq_compress(&list);

  // Print.

  if (label)
  {
    print(str::Format("%1:").arg(in));
    if (_verbose) separator();
  }
  
  for (uint i = 0; i < list.size(); ++i)
  {
    try { file(list[i]); }
    catch (Error) {}
  }
  
  separator();
  
  // Recurse.
  
  if (_recurse)
  {
    Dir dir(in);
    List<File> list = dir.list();
    file::filter(&list, file::FILTER_FILE | file::FILTER_HIDDEN);
    for (uint i = 0; i < list.size(); ++i)
      this->dir(list[i], label);
  }
}

Str App::label_time(int in, const image::time::Speed & speed) const
{
  return _timecode ?
    image::time::timecode_to_str(image::time::frame_to_timecode(in, speed)) :
    str::int_to_str(in);
}

void App::cmdln() throw (Error)
{
  if (! cmdln_list.size()) { exit(app::EXIT_HELP); return; }

  while (cmdln_list.size())
  {
    const Str arg = cmdln_list.pop_front();
    
    // Options.
    
    if (str::compare("-verbose", arg) || str::compare("-v", arg))
      _verbose = true;
    else if (str::compare("-x_seq", arg) || str::compare("-q", arg))
      _seq = false;
    else if (str::compare("-recurse", arg) || str::compare("-r", arg))
      _recurse = true;
    else if (str::compare("-x_timecode", arg) || str::compare("-t", arg))
      _timecode = false;
    else if (str::compare("-columns", arg) || str::compare("-c", arg))
    {
      if (! str::serialize(&cmdln_list, &_columns)) throw Error(arg);
    }
    
    // Load options.
    
    else if (str::compare("-load", arg))
    {
      Str plugin, option;
      if (! str::serialize(&cmdln_list, &plugin)) throw Error(arg);
      if (! str::serialize(&cmdln_list, &option)) throw Error(arg);
      if (! image::io::load()->option(plugin, option, &cmdln_list))
        throw Error(arg);
    }
    else if (str::compare("-load_plugin", arg))
    {
      if (! str::serialize(&cmdln_list, &_load)) throw Error(arg);
    }
    
    // Arguments.
    
    else _input += arg;
  }

  if (! _input.size()) throw Error("input");
}

const Str _help =
"\n"
"djv_info\n"
"--------\n"
"\n"
"Image information utility.\n"
"\n"
"Output:\n"
"\n"
"yesterdayview.mov                         640x424:1.51 RGB U8 00:02:00:01@12\n"
"dlad.dpx                               2048x1556:1.32 RGB U10 00:00:00:01@24\n"
"render0001-1000.exr                       720x480:1.5 RGB F16 00:00:41:16@24\n"
"#1 Ambient                                               720x480:1.5 RGB F16\n"
"#2 Diffuse                                               720x480:1.5 RGB F16\n"
"#3 Specular                                              720x480:1.5 RGB F16\n"
"#4 Z                                                       720x480:1.5 L F32\n"
"\n"
"Image Format:\n"
"\n"
"(name)          (width)x(height):(aspect) (format) (type) (duration):(speed)\n"
"\n"
"Layer Format:\n"
"\n"
"#(layer) (name)                    (width)x(height):(aspect) (format) (type)\n"
"\n"
"Command Line\n"
"------------\n"
"\n"
"Usage:\n"
"\n"
"djv_info (image|directory)... [option]...\n"
"\n"
"Options:\n"
"\n"
"-verbose, -v - Verbose information.\n"
"-x_seq, -q - Disable file sequencing.\n"
"-recurse, -r - Descend into sub-directories.\n"
"-x_timecode, -t - Disable timecode display.\n"
"-columns, -c (value) - Number of output columns. Zero disables formatting.\n"
"\n"
"Load Options:\n"
"\n"
"-load (plugin) (option) (value) - Set a plugin option:\n"
"%1"
"\n"
"-load_plugin (value) - Specify load plugin: %2.\n"
"\n"
"%3"
"\n"
"Examples\n"
"--------\n"
"\n"
"Image information:\n"
"> djv_info image.sgi image2.sgi\n"
"\n"
"Image sequence information:\n"
"> djv_info image.1-100.sgi\n"
"\n"
"Information about all images within a directory:\n"
"> djv_info ~/pics\n"
"\n"
"Video lookup table information:\n"
"> djv_info 0.vlut\n";

Str App::help() const
{
  return str::Format(_help).
    arg(image::io::load()->help()).
    arg(str::join(image::io::load()->name_list(), ", ")).
    arg(image::App::help());
}

}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = ::App(argc, argv).exec(); }
  catch (djv::Error in) { djv::error::print(in); }
  return r;
}

