/*------------------------------------------------------------------------------
 bin/djv_view/playback_widget.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "playback.h"
#include <djv_image/time.h>
#include <djv_ui/style.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/text.h>
#include <djv_ui/menu.h>
#include <FL/fl_draw.H>

namespace djv_view {
namespace playback
{

Frame_Widget::Frame_Widget() : _frame(0)
{
  // Widgets.
  
  _widget = new ui::text::Edit;
  
  ui::num::Inc * inc_widget = new ui::num::Inc;
  
  // Layout.
  
  Item::size(ui::layout::FIXED, ui::layout::FIXED);
  Item::align(ui::layout::LEFT, ui::layout::CENTER);
  
  ui::layout::Horizontal * layout = new ui::layout::Horizontal(this);
  layout->spacing(2);
  layout->margin(0);
  layout->add(_widget);
  layout->stretch(_widget);
  layout->add(inc_widget);
  
  // Initialize.
  
  value_update();
  
  // Callbacks.
  
  _widget->sig.set(this, widget_callback);
  inc_widget->inc_sig.set(this, inc_callback);
  inc_widget->dec_sig.set(this, dec_callback);
}

void Frame_Widget::speed(const image::time::Speed & in)
{
  if (in == _speed) return;
  _speed = in;
  value_update();
}

void Frame_Widget::text_font(const Fl_Font & in)
{
  Widget::text_font(in);
  _widget->text_font(in);
}

void Frame_Widget::text_color(const Fl_Color & in)
{
  Widget::text_color(in);
  _widget->text_color(in);
}
  
int Frame_Widget::handle(int in)
{
  if (Int_Base::handle(in)) return 1;
  return Widget::handle(in);
}

void Frame_Widget::dirty()
{
  DJV_DBG_OBJ("playback::Frame_Widget::dirty");

  _widget->size_str(ui::style::global()->timecode() ? "00:00:00:00" : "000000");

  Widget::dirty();
  
  value_update();
  
  DJV_DBG_PRN("hint = " << size_hint());
}

ui::layout::Item * Frame_Widget::group_align() { return _widget; }

void Frame_Widget::widget_callback(const Str & in)
{
  DJV_DBG_OBJ("playback::Frame_Widget::widget_callback");
  DJV_DBG_PRN("in = " << in);
  
  Int_Base::set(ui::style::global()->timecode() ?
    image::time::timecode_to_frame(
      image::time::str_to_timecode(in), _speed) :
    str::str_to_int(in));
  
  value_update();
}

void Frame_Widget::inc_callback(bool) { Int_Base::inc_small_action(); }
void Frame_Widget::dec_callback(bool) { Int_Base::dec_small_action(); }

void Frame_Widget::value_update()
{
  callbacks(false);
  
  _widget->set(ui::style::global()->timecode() ?
    image::time::timecode_to_str(
      image::time::frame_to_timecode(get(), _speed)) :
    str::int_to_str(get()));
  
  callbacks(true);
}

Frame_Slider::Frame_Slider() : _in(0), _out(0) {}

void Frame_Slider::in_out(int in, int out)
{
  _in = in;
  _out = out;
  redraw();
}

void Frame_Slider::draw()
{
  Frame::draw();

  if (_in > min() || _out < max() && min() != max())
  {
    const Box2i box = frame_geom();

    const double range = max() - min();
    const double v0 = (math::max(_in, min()) - min()) / range;
    const double v1 = (math::min(_out, max()) - min()) / range;

    fl_color(FL_SELECTION_COLOR);
    fl_rectf(int(box.x + box.w * v0), box.y, int(box.w * (v1 - v0) + 1), box.h);
  }
  
  handle_draw(min() != max() ? ((get() - min()) / double(max() - min())) : 0,
    frame_geom());
}

Time_Widget::Time_Widget() : _frame(0), _mark(false)
{
  // Widgets.
  
  _widget = new ui::text::Display;
  
  _mark_widget = new Frame_Widget;
  _mark_widget->text_color(FL_SELECTION_COLOR);
  
  // Layout.
  
  Item::size(ui::layout::FIXED, ui::layout::FIXED);
  Item::align(ui::layout::LEFT, ui::layout::CENTER);
  
  ui::layout::Horizontal * layout = new ui::layout::Horizontal(this);
  layout->margin(0);
  layout->add(_widget);
  layout->add(_mark_widget);
  
  // Initialize.

  value_update();
  widget_update();
  
  // Callbacks.
  
  _mark_widget->sig.set(this, widget_callback);
}

void Time_Widget::frame(int in)
{
  if (in == _frame) return;
  _frame = in;
  value_update();
}

void Time_Widget::speed(const image::time::Speed & in)
{
  if (in == _speed) return;
  _speed = in;
  value_update();
}

void Time_Widget::mark(bool in)
{
  if (in == _mark) return;
  _mark = in;
  widget_update();
}

void Time_Widget::range(int start, int end)
{
  _mark_widget->range(start, end);
}

void Time_Widget::dirty()
{
  _widget->size_str(ui::style::global()->timecode() ? "00:00:00:00" : "000000");

  Widget::dirty();

  value_update();
}

void Time_Widget::widget_callback(int in)
{
  mark_sig.emit(in);
}

void Time_Widget::value_update()
{
  callbacks(false);

  if (_mark)
  {
    _mark_widget->set(_frame);
    _mark_widget->speed(_speed);
  }
  else
  {
    _widget->set(ui::style::global()->timecode() ?
      image::time::timecode_to_str(
        image::time::frame_to_timecode(_frame, _speed)) :
      str::int_to_str(_frame));
  }

  callbacks(true);
}

void Time_Widget::widget_update()
{
  callbacks(false);
  
  if (_mark)
  {
    _widget->hide();
    _mark_widget->show();
  }
  else
  {
    _mark_widget->hide();
    _widget->show();
  }
  
  update();

  callbacks(true);
}

Speed_Widget::Speed_Widget()
{
  _widget = new ui::text::Display;
  _menu = new ui::menu::Popup;

  Item::size(ui::layout::FIXED, ui::layout::FIXED);
  
  ui::layout::Horizontal * layout = new ui::layout::Horizontal(this);
  layout->margin(0);
  layout->add(_widget);
  layout->stretch(_widget);
  layout->add(_menu);

  widget_update();
  menu_update();
  
  _menu->sig.set(this, menu_callback);
}

void Speed_Widget::set(const image::time::Speed & in)
{
  if (in == _value) return;
  _value = in;
  widget_update();
  sig.emit(_value);
}

void Speed_Widget::set_default(const image::time::Speed & in)
{
  if (in == _value_default) return;
  _value_default = in;
  menu_update();
}

void Speed_Widget::dirty()
{
  _widget->size_str("000.00");

  Widget::dirty();
}

void Speed_Widget::menu_callback(int in)
{
  if (_value_default.valid())
    set(0 == in ? _value_default : image::time::FPS(in - 1));
  else
    set(image::time::FPS(in));
}

void Speed_Widget::widget_update()
{
  callbacks(false);
  _widget->set(str::float_to_str(image::time::speed_to_float(_value), -2));
  callbacks(true);
}

void Speed_Widget::menu_update()
{
  callbacks(false);
  
  _menu->menu_del();
  if (_value_default.valid())
    _menu->menu_add(str::float_to_str(image::time::speed_to_float(
      _value_default), -2), 0, 0, 0, ui::menu::DIVIDER);
  _menu->add(image::time::label_fps());
  _menu->menu_end();
  
  callbacks(true);
}

Speed_Display_Widget::Speed_Display_Widget() :
  _value(0.0), _dropped_frames(false)
{
  _widget = new ui::text::Display;

  Item::size(ui::layout::FIXED, ui::layout::FIXED);
  
  ui::layout::Horizontal * layout = new ui::layout::Horizontal(this);
  layout->margin(0);
  layout->add(_widget);

  widget_update();
}

void Speed_Display_Widget::set(double in)
{
  if (in == _value) return;
  _value = in;
  widget_update();
}

void Speed_Display_Widget::dropped_frames(bool in)
{
  if (in == _dropped_frames) return;
  _dropped_frames = in;
  _widget->text_color(_dropped_frames ? FL_SELECTION_COLOR :
    FL_FOREGROUND_COLOR);
  redraw();
}

void Speed_Display_Widget::dirty()
{
  _widget->size_str("000.00");

  Widget::dirty();
}

void Speed_Display_Widget::widget_update()
{
  callbacks(false);
  _widget->set(str::float_to_str(_value, -2));
  callbacks(true);
}

}
}

