/*------------------------------------------------------------------------------
 bin/djv_view/window.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "window.h"
#include <djv_ui/style.h>
#include <djv_ui/icon.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/group.h>
#include <djv_ui/button.h>
#include <djv_ui/menu.h>
#include <djv_ui/dialog.h>
#include <djv_ui/image_io.h>
#include <djv_ui/app.h>
#include "prefs.h"
#include "image_color.h"
#include "magnify.h"
#include "color_sample.h"
#include "histogram.h"
#include "image_info.h"
#include <FL/Fl.H>

namespace djv_view {
namespace window
{

Window_Group::Window_Group() : _menu(0) {}

void Window_Group::fullscreen(bool in) { _fullscreen_widget->set(in); }

void Window_Group::menu(ui::menu::Widget * in, bool fullscreen, bool controls)
{
  in->menu_add("&Window", 0, 0, 0, ui::menu::SUB_MENU);
  
  in->menu_add("&New", FL_ALT + 'n', _new_callback, this);
  in->menu_add("&Copy", FL_ALT + 'c', _copy_callback, this);
  in->menu_add("Clos&e", FL_ALT + 'w', _close_callback, this,
    ui::menu::DIVIDER);

  in->menu_add("&Fit", 'f', _fit_callback, this);
  in->menu_add("F&ullscreen", 'u', _fullscreen_callback, this,
    ui::menu::TOGGLE | (fullscreen ? ui::menu::VALUE : 0));
  in->menu_add("Contro&ls", 'l', _controls_callback, this,
    ui::menu::TOGGLE | (controls ? ui::menu::VALUE : 0));

  in->menu_end();
  
  _menu = in;
}

void Window_Group::toolbar(ui::layout::Horizontal * in)
{
  ui::button::Tool * new_widget = new ui::button::Tool("window_new");
  new_widget->tooltip("Create a new window.\nKeyboard shortcut: Alt + N.");
  ui::button::Tool * copy_widget = new ui::button::Tool("window_copy");
  copy_widget->tooltip("Copy the window.\nKeyboard shortcut: Alt + C.");
  ui::button::Tool * close_widget = new ui::button::Tool("window_close");
  close_widget->tooltip("Close the window.\nKeyboard shortcut: Alt + W.");
  ui::button::Tool * fit_widget = new ui::button::Tool("window_fit");
  fit_widget->tooltip("Fit window to image.\nKeyboard shortcut: F.");
  _fullscreen_widget = new ui::button::Tool("window_fullscreen");
  _fullscreen_widget->type(ui::button::Tool::TOGGLE);
  _fullscreen_widget->tooltip(
    "Toggle fullscreen display.\nKeyboard shortcut: U.");
  
  in->add(new_widget);
  in->add(copy_widget);
  in->add(close_widget);
  in->add(fit_widget);
  in->add(_fullscreen_widget);
  
  new_widget->sig.set(this, new_callback);
  copy_widget->sig.set(this, copy_callback);
  close_widget->sig.set(this, close_callback);
  fit_widget->sig.set(this, fit_callback);
  _fullscreen_widget->sig.set(this, fullscreen_callback);
}

void Window_Group::new_callback(bool) { factory()->add()->show(); }
void Window_Group::_new_callback() { new_callback(true); }
void Window_Group::copy_callback(bool) { copy_sig.emit(true); }
void Window_Group::_copy_callback() { copy_callback(true); }
void Window_Group::close_callback(bool) { close_sig.emit(true); }
void Window_Group::_close_callback() { close_callback(true); }
void Window_Group::fit_callback(bool) { fit_sig.emit(true); }
void Window_Group::_fit_callback() { fit_callback(true); }
void Window_Group::fullscreen_callback(bool in) { fullscreen_sig.emit(in); }
void Window_Group::_fullscreen_callback()
{ fullscreen_callback(_menu->value()); }
void Window_Group::_controls_callback() { controls_sig.emit(_menu->value()); }

Widget::Widget(Widget * copy) :
  ui::window::Widget(ui::app::global()->name()),
  _file(copy ? copy->_file : File()),
  _load(0),
  _file_layer(copy ? copy->_file_layer : 0),
  _file_proxy(copy ? copy->_file_proxy : prefs::file()->proxy()),
  _file_cache(copy ? copy->_file_cache : prefs::file()->cache()),
  _window_controls(true),
  _view_grid(copy ? copy->_view_grid : prefs::view()->grid()),
  _view_hud(copy ? copy->_view_hud : prefs::view()->hud()),
  _image_p(0), _image_ref(0),
  _image_mirror_x(copy ? copy->_image_mirror_x : prefs::image()->mirror_x()),
  _image_mirror_y(copy ? copy->_image_mirror_y : prefs::image()->mirror_y()),
  _image_scale(copy ? copy->_image_scale : prefs::image()->scale()),
  _image_rotate(copy ? copy->_image_rotate : prefs::image()->rotate()),
  _image_profile(copy ? copy->_image_profile : prefs::image()->profile()),
  _image_color(copy ? copy->_image_color : prefs::image()->color_value()),
  _image_channel(copy ? copy->_image_channel : prefs::image()->channel()),
  _image_filter(copy ? copy->_image_filter : prefs::image()->filter()),
  _playback_mode(copy ? copy->_playback_mode : playback::STOP),
  _playback_time(copy ? copy->_playback_time : image::Time()),
  _playback_frame(copy ? copy->_playback_frame : 0), _playback_frame_tmp(0),
  _playback_in(copy ? copy->_playback_in : 0),
  _playback_out(copy ? copy->_playback_out : 0),
  _playback_every_frame(
    copy ? copy->_playback_every_frame : prefs::playback()->every_frame()),
  _playback_shuttle(false), _playback_shuttle_speed(0.0),
  _playback_speed_real(0.0), _playback_dropped_frames(false),
  _playback_dropped_frames_tmp(false),
  _playback_timer_init(true), _playback_timer_frame(0),
  _playback_speed_counter(0)
{
  DJV_DBG_OBJ(str::Format("window::Widget::Widget(%1)").arg(int(this)));
  
  // Widgets.

  _file_group = std::auto_ptr<File_Group>(new File_Group);
  _window_group = std::auto_ptr<Window_Group>(new Window_Group);
  _view_group = std::auto_ptr<View_Group>(new View_Group);
  _image_group = std::auto_ptr<Image_Group>(new Image_Group);
  _playback_group = std::auto_ptr<Playback_Group>(new Playback_Group);

  _menu = new ui::menu::Widget;
  _toolbar = new ui::group::Frame;
  _toolbar->edges(false, false, true, true);
  _playbar = new ui::group::Frame;
  _playbar->edges(false, false, true, false);
  
  _view = new view::Widget;

  // Layout.

  ui::layout::Vertical * layout = new ui::layout::Vertical(this);
  layout->spacing(0);
  layout->margin(0);
  layout->add(_menu);
  
  layout->add(_toolbar);
  ui::layout::Horizontal * toolbar_layout = new ui::layout::Horizontal(_toolbar);
  toolbar_layout->spacing(0);
  toolbar_layout->margin(0);
  _file_group->toolbar(toolbar_layout);
  toolbar_layout->add(
    new ui::group::Horizontal_Divider(ui::style::global()->margin()));
  _window_group->toolbar(toolbar_layout);
  toolbar_layout->add(
    new ui::group::Horizontal_Divider(ui::style::global()->margin()));
  _view_group->toolbar(toolbar_layout);
  toolbar_layout->add(
    new ui::group::Horizontal_Divider(ui::style::global()->margin()));
  _image_group->toolbar(toolbar_layout);
  toolbar_layout->add_stretch();
  
  layout->add(_view);
  layout->stretch(_view);

  _playback_group->playbar(_playbar);
  layout->add(_playbar);

  // Initialize.
  
  window_controls_update();
  view_update();
  view_overlay_update();
  playback_update();
  playback_time_update();
  playback_speed_update();
  playback_frame_update();
  menu_update();
  
  if (copy)
  {
    file_open(_file, true);
    _view->view(copy->_view->view());
    _view->zoom(copy->_view->zoom());
    size(copy->geom().size);
  }
  else
  {
    window_fit();
  }
  
  // Callbacks.
  
  _file_group->open_sig.set(this, file_open_callback);
  _file_group->reload_sig.set(this, file_reload_callback);
  _file_group->close_sig.set(this, file_close_callback);
  _file_group->layer_sig.set(this, file_layer_callback);
  _file_group->layer_prev_sig.set(this, file_layer_prev_callback);
  _file_group->layer_next_sig.set(this, file_layer_next_callback);
  _file_group->proxy_sig.set(this, file_proxy_callback);
  _file_group->cache_sig.set(this, file_cache_callback);
  prefs::file()->recent_sig.set(this, file_recent_callback);
  prefs::file()->proxy_sig.set(this, file_proxy_callback);
  prefs::file()->cache_sig.set(this, file_cache_callback);
  ui::image_io::sig.set(this, file_reload_callback);
  
  _window_group->copy_sig.set(this, window_copy_callback);
  _window_group->close_sig.set(this, window_close_callback);
  _window_group->fit_sig.set(this, window_fit_callback);
  _window_group->fullscreen_sig.set(this, window_fullscreen_callback);
  _window_group->controls_sig.set(this, window_controls_callback);
  prefs::window()->controls_sig.set(this, window_controls_callback);

  _view_group->move_sig.set(this, view_move_callback);
  _view_group->center_sig.set(this, view_center_callback);
  _view_group->zoom_sig.set(this, view_zoom_callback);
  _view_group->zoom_reset_sig.set(this, view_zoom_reset_callback);
  _view_group->reset_sig.set(this, view_reset_callback);
  _view_group->fit_sig.set(this, view_fit_callback);
  _view_group->grid_sig.set(this, view_grid_callback);
  _view_group->hud_sig.set(this, view_hud_callback);
  _view->sig.set(this, view_callback);
  _view->pick_sig.set(this, view_pick_callback);
  _view->file_sig.set(this, file_open_callback);
  prefs::view()->grid_sig.set(this, view_grid_callback);
  prefs::view()->grid_color_sig.set(this, view_grid_color_callback);
  prefs::view()->hud_sig.set(this, view_hud_callback);
  prefs::view()->hud_foreground_sig.set(this, view_hud_foreground_callback);
  prefs::view()->hud_background_sig.set(this, view_hud_background_callback);
  prefs::view()->background_sig.set(this, view_background_callback);
  
  _image_group->mirror_x_sig.set(this, image_mirror_x_callback);
  _image_group->mirror_y_sig.set(this, image_mirror_y_callback);
  _image_group->scale_sig.set(this, image_scale_callback);
  _image_group->rotate_sig.set(this, image_rotate_callback);
  _image_group->profile_sig.set(this, image_profile_callback);
  _image_group->color_sig.set(this, image_color_callback);
  _image_group->color_value_sig.set(this, image_color_callback);
  _image_group->color_reset_sig.set(this, image_color_reset_callback);
  _image_group->color_add_sig.set(this, image_color_add_callback);
  _image_group->channel_sig.set(this, image_channel_callback);
  _image_group->filter_sig.set(this, image_filter_callback);
  _image_group->magnify_sig.set(this, image_magnify_callback);
  _image_group->color_sample_sig.set(this, image_color_sample_callback);
  _image_group->histogram_sig.set(this, image_histogram_callback);
  _image_group->info_sig.set(this, image_info_callback);
  prefs::image()->mirror_x_sig.set(this, image_mirror_x_callback);
  prefs::image()->mirror_y_sig.set(this, image_mirror_y_callback);
  prefs::image()->scale_sig.set(this, image_scale_callback);
  prefs::image()->rotate_sig.set(this, image_rotate_callback);
  prefs::image()->profile_sig.set(this, image_profile_callback);
  prefs::image()->color_value_sig.set(this, image_color_callback);
  prefs::image()->color_list_sig.set(this, image_color_list_callback);
  prefs::image()->channel_sig.set(this, image_channel_callback);
  prefs::image()->filter_sig.set(this, image_filter_callback);

  _playback_group->playback_sig.set(this, playback_callback);
  _playback_group->playback_shuttle_sig.set(this, playback_shuttle_callback);
  _playback_group->playback_shuttle_value_sig.set(
    this, playback_shuttle_value_callback);
  _playback_group->frame_sig.set(this, playback_frame_callback);
  _playback_group->frame_move_sig.set(this, playback_frame_move_callback);
  _playback_group->frame_shuttle_sig.set(this, playback_frame_shuttle_callback);
  _playback_group->frame_shuttle_value_sig.set(
    this, playback_frame_shuttle_value_callback);
  _playback_group->in_sig.set(this, playback_in_callback);
  _playback_group->in_mark_sig.set(this, playback_in_mark_callback);
  _playback_group->in_reset_sig.set(this, playback_in_reset_callback);
  _playback_group->out_sig.set(this, playback_out_callback);
  _playback_group->out_mark_sig.set(this, playback_out_mark_callback);
  _playback_group->out_reset_sig.set(this, playback_out_reset_callback);  
  _playback_group->speed_sig.set(this, playback_speed_callback);
  _playback_group->every_frame_sig.set(this, playback_every_frame_callback);
  prefs::playback()->every_frame_sig.set(this, playback_every_frame_callback);
  
  close_sig.set(this, window_close2_callback);
}

Widget::~Widget()
{
  DJV_DBG_OBJ(str::Format("window::Widget::~Widget(%1)").arg(int(this)));
}

void Widget::window_fit(bool init)
{
  DJV_DBG_OBJ("window::Widget::window_fit");
  DJV_DBG_PRN("init = " << init);
  DJV_DBG_PRN("info = " << _info.size());
  
  Vec2i image_size;
  if (! init) image_size = _view->screen_size();
  if (! vec::size_valid(image_size) && _info.size())
  {
    // Fit to screen size.
      
    if (vec::size_valid(_info[_file_layer].size))
    {
      image_size = _info[_file_layer].size;
      const Vec2i max(int(Fl::w() * 0.75), int(Fl::h() * 0.75));
      if (image_size.x > max.x || image_size.y > max.y)
        image_size = (image_size.x > image_size.y) ?
          Vec2i(max.x, int(image_size.y / double(image_size.x) * max.x)) :
          Vec2i(int(image_size.x / double(image_size.y) * max.y), max.y);
    }
  }
  if (! vec::size_valid(image_size))
  {
    // Default size.
      
    image_size = Vec2i(600, 300);
  }
  DJV_DBG_PRN("image size = " << image_size);

  const Vec2i ui_size = size_hint() - Vec2i(0, _view->size_hint().y);
  DJV_DBG_PRN("ui size = " << ui_size);
  
  const Vec2i size = Vec2i(
    math::max(image_size.x, ui_size.x),
    image_size.y + ui_size.y);
  DJV_DBG_PRN("size = " << size);
  
  this->size(size);
  window_fullscreen_update();
  _view->view_fit();
  _view->redraw();
}

void Widget::widget_del()
{
  DJV_DBG_OBJ(str::Format("window::Widget::widget_del(%1)").arg(int(this)));
  
  Fl::remove_idle(playback_timer_callback, this);
  //Fl::remove_timeout(playback_timer_callback, this);
  
  magnify::dialog()->del(_view);
  color_sample::dialog()->del(_view);
  histogram::dialog()->del(_view);
  image_info::dialog()->del(this);
  
  _image_color_sig.del();
  
  _file_group->del();
  _window_group->del();
  _view_group->del();
  _image_group->del();
  _playback_group->del();
  
  file_cache_update();

  _view->set(0);
  if (_image_ref) _image_ref->ref_del();
  
  ui::window::Widget::widget_del();
}

void Widget::window_copy_callback(bool) { factory()->add(this)->show(); }

void Widget::window_close_callback(bool)
{
  if (1 == factory()->list().size()) return;
  window_close2_callback(true);
}

void Widget::window_close2_callback(bool) { factory()->del(this); }

void Widget::window_fit_callback(bool) { window_fit(); }

void Widget::window_fullscreen_callback(bool in)
{
  fullscreen(in);
  window_fullscreen_update();
  menu_update();
}

void Widget::window_controls_callback(bool in)
{
  _window_controls = in;
  window_controls_update();
  menu_update();
}

void Widget::window_controls_update()
{
  callbacks(false);
  
  if (_window_controls)
  {
    _toolbar->show();
    _playbar->show();
  }
  else
  {
    _toolbar->hide();
    _playbar->hide();
  }

  update();

  callbacks(true);
}

void Widget::window_fullscreen_update()
{
  callbacks(false);
  
  _window_group->fullscreen(fullscreen());
  if (prefs::window()->resize()) _view->view_fit();
  
  callbacks(true);
}

void Widget::menu_update()
{
  callbacks(false);
  
  _menu->menu_del();
  _file_group->menu(_menu, _file_layer, _file_layer_list, _file_proxy,
    _file_cache);
  _window_group->menu(_menu, fullscreen(), _window_controls);
  _view_group->menu(_menu, _view_grid, _view_hud);
  _image_group->menu(_menu, _image_mirror_x, _image_mirror_y, _image_scale,
    _image_rotate, _image_profile, _image_channel, _image_filter);
  _playback_group->menu(_menu, _playback_mode, _playback_every_frame);
  _menu->menu_end();
  
  callbacks(true);
}

}
}

