/*------------------------------------------------------------------------------
 lib/djv_base/box.h
 
 Boxes
 -----
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_BOX_H
#define DJV_BOX_H

#include <djv_base/vec.h>

namespace djv {
namespace box
{

//------------------------------------------------------------------------------
// Box Base

template<typename T, int D>
class Box
{
public:

  Box() {}
  Box(const Box<T, D> & in) { set(in); }
  Box(const Vec<T, D> & position, const Vec<T, D> & size) :
    position(position), size(size)
  {}
  Box(const Vec<T, D> & size) : size(size) {}

  Vec<T, D> position, size;
  
  void set(const Box<T, D> & in) { position = in.position; size = in.size; }
  void zero() { position.zero(); size.zero(); }
};

//------------------------------------------------------------------------------
// Two-Dimensional Box

template<typename T>
class Box2 : public Box<T, 2>
{
public:

  inline Box2();
  inline Box2(const Box2<T> &);
  inline Box2(const Box<T, 2> &);
  inline Box2(const Vec<T, 2> & position, const Vec<T, 2> & size);
  inline Box2(const Vec<T, 2> & size);
  inline Box2(T x, T y, T w, T h);
  inline Box2(T w, T h);
  
  T & x, & y, & w, & h;
  
  Box2<T> & operator = (const Box2<T> & in) { set(in); return *this; }

  inline void position1(const Vec<T, 2> &);
  inline Vec<T, 2> position1() const;
  inline void x1(const T &);
  inline void y1(const T &);
  inline T x1() const;
  inline T y1() const;
};

//------------------------------------------------------------------------------
// Three-Dimensional Box

template<typename T>
class Box3 : public Box<T, 3>
{
public:

  inline Box3();
  inline Box3(const Box3<T> &);
  inline Box3(const Box<T, 3> &);
  inline Box3(const Vec<T, 3> & position, const Vec<T, 3> & size);
  inline Box3(const Vec<T, 3> & size);
  inline Box3(T x, T y, T z, T w, T h, T d);
  inline Box3(T w, T h, T d);

  T & x, & y, & z, & w, & h, & d;
  
  Box3<T> & operator = (const Box3<T> & in) { set(in); return *this; }

  inline void position1(const Vec<T, 3> &);
  inline Vec<T, 3> position1() const;
  inline void x1(const T &);
  inline void y1(const T &);
  inline void z1(const T &);
  inline T x1() const;
  inline T y1() const;
  inline T z1() const;
};

//------------------------------------------------------------------------------
// Functions

template<typename T, int D>
inline Box<T, D> swap(const Box<T, D> &);

template<typename T>
inline bool intersect(const Box2<T> &, const Vec2<T> &);
template<typename T>
inline Box2<T> intersect(const Box2<T> &, const Box2<T> &);
template<typename T>
inline bool intersect(const Box3<T> &, const Vec3<T> &);
template<typename T>
inline Box3<T> intersect(const Box3<T> &, const Box3<T> &);

template<typename T>
inline Box2<T> bound(const Box2<T> &, const Box2<T> &);
template<typename T>
inline Box3<T> bound(const Box3<T> &, const Box3<T> &);

template<typename T>
inline Box2<T> expand(const Box2<T> &, const Vec2<T> &);
template<typename T>
inline Box3<T> expand(const Box3<T> &, const Vec3<T> &);

template<typename T, int D>
inline bool size_valid(const Box<T, D> &);

//------------------------------------------------------------------------------
// Operators

template <typename T, int D>
inline Box<T, D> operator + (const Box<T, D> &, T);
template <typename T, int D>
inline Box<T, D> operator - (const Box<T, D> &, T);
template <typename T, int D>
inline Box<T, D> operator * (const Box<T, D> &, T);
template <typename T, int D>
inline Box<T, D> operator / (const Box<T, D> &, T);

//------------------------------------------------------------------------------
// Comparison

template <typename T, int D>
inline bool operator == (const Box<T, D> &, const Box<T, D> &);
template <typename T, int D>
inline bool operator != (const Box<T, D> &, const Box<T, D> &);

//------------------------------------------------------------------------------
// Conversion

template <int D>
inline Box<int, D> to_int(const Box<double, D> &);
template <int D>
inline Box<double, D> to_float(const Box<int, D> &);

template <typename T, int D>
inline bool serialize(List<Str> *, Box<T, D> *);
template <typename T, int D>
inline List<Str> serialize(const Box<T, D> &);

//------------------------------------------------------------------------------
// Debugging

template <typename T, int D>
inline Dbg & operator << (Dbg &, const Box<T, D> &);

}
}

#include <djv_base/box2_.h>
#include <djv_base/box3_.h>
#include <djv_base/box_fnc_.h>

#endif

