/*------------------------------------------------------------------------------
 lib/djv_base/mtx_fnc_.h

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include <djv_base/math.h>

namespace djv {
namespace mtx
{

template<typename T, int E>
inline Mtx<T, E> transpose(const Mtx<T, E> & in)
{
  Mtx<T, E> out;
  for (int i = 0; i < E; ++i)
    for (int j = 0; j < E; ++j)
      out.e[j * E + i] = in.e[i * E + j];
  return out;
}

template<typename T, int E, int V>
inline Mtx<T, E> translate(const Vec<T, V> & in)
{
  //DJV_ASSERT((V + 1) == E);
  Mtx<T, E> out;
  for (int i = 0; i < V; ++i)
    out.e[(E - 1) * E + i] = in.e[i];
  return out;
}

inline Mtx3f translate3f(const Vec2f & in)
{ return translate<double, 3, 2>(in); }

template<typename T, int E, int V>
inline Mtx<T, E> scale(const Vec<T, V> & in)
{
  //DJV_ASSERT((V + 1) == E);
  Mtx<T, E> out;
  for (int i = 0; i < V; ++i)
    out.e[i * E + i] = in.e[i];
  return out;
}

inline Mtx3f scale3f(const Vec2f & in)
{ return scale<double, 3, 2>(in); }

inline Mtx3f rotate3f(double in)
{
  const double cos = math::cos(math::degrees_to_radians(in));
  const double sin = math::sin(math::degrees_to_radians(in));
  return Mtx3f(
    cos, sin, 0,
    -sin, cos, 0,
    0, 0, 1);
}

template<typename T, int E>
inline Mtx<T, E> operator * (const Mtx<T, E> & a, const Mtx<T, E> & b)
{
  Mtx<T, E> out;
  out.zero();
  for (int i = 0; i < E; ++i)
    for (int j = 0; j < E; ++j)
      for (int k = 0; k < E; ++k)
        out.e[i * E + j] += a.e[i * E + k] * b.e[k * E + j];
  return out;
}

template<typename T, int E, int V>
inline Vec<T, V> operator * (const Mtx<T, E> & in, const Vec<T, V> & v)
{
  //DJV_ASSERT((V + 1) == E);
  Vec<T, V> out;
  for (int i = 0; i < V; ++i)
  {
    int j = 0;
    for (; j < V; ++j)
      out.e[i] += in.e[j * E + i] * v.e[j];
    out.e[i] += in.e[j * E + i];
  }
  return out;
}

template<typename T, int E>
inline bool operator == (const Mtx<T, E> & a, const Mtx<T, E> & b)
{
  for (int i = 0; i < E * E; ++i)
    if (a.e[i] != b.e[i]) return false;
  return true;
}

template<typename T, int E>
inline bool operator != (const Mtx<T, E> & a, const Mtx<T, E> & b)
{ return ! (a == b); }

template<typename T, int E>
inline bool serialize(List<Str> * in, Mtx<T, E> * out)
{
  for (int i = 0; i < E * E; ++i)
    if (! str::serialize(in, &out->e[i])) return false;
  return true;
}

template<typename T, int E>
inline List<Str> serialize(const Mtx<T, E> & in)
{
  List<Str> out;
  for (int i = 0; i < E * E; ++i)
    out += str::serialize(in.e[i]);
  return out;
}

template<typename T, int E>
inline Dbg & operator << (Dbg & dbg, const Mtx<T, E> & in)
{
  dbg << "\n";
  for (int i = 0; i < E; ++i) {
    for (int j = 0; j < E; ++j) {
      dbg << in.e[i * E + j];
      if (j < E - 1) dbg << " ";
    }
    if (i < E - 1) dbg << "\n";
  }
  return dbg;
}

}
}

