/*------------------------------------------------------------------------------
 lib/djv_base/vec.h
 
 Vectors
 -------

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_VEC_H
#define DJV_VEC_H

#include <djv_base/base.h>

namespace djv {
namespace vec
{

//------------------------------------------------------------------------------
// Vector Base

template<typename T, int D>
class Vec
{
public:

  Vec() { zero(); }
  Vec(const Vec<T, D> & in) { set(in); }
  Vec(T in) { set(in); }
  virtual ~Vec() {}
  
  T e [D];
  
  inline void set(const Vec<T, D> &);
  inline void set(T);
  inline void zero();
  Vec<T, D> & operator = (const Vec<T, D> &);
  
  inline Vec<T, D> & operator += (const Vec<T, D> &);
  inline Vec<T, D> & operator -= (const Vec<T, D> &);
  inline Vec<T, D> & operator *= (const Vec<T, D> &);
  inline Vec<T, D> & operator /= (const Vec<T, D> &);
  inline Vec<T, D> & operator += (T);
  inline Vec<T, D> & operator -= (T);
  inline Vec<T, D> & operator *= (T);
  inline Vec<T, D> & operator /= (T);
};

//------------------------------------------------------------------------------
// Two-Dimensional Vector

template<typename T>
class Vec2 : public Vec<T, 2>
{
public:

  inline Vec2();
  inline Vec2(const Vec2<T> &);
  inline Vec2(const Vec<T, 2> &);
  inline Vec2(T, T);
  inline Vec2(T);
  
  T & x, & y;
  
  Vec2<T> & operator = (const Vec2<T> & in) { set(in); return *this; }
};

//------------------------------------------------------------------------------
// Three-Dimensional Vector

template<typename T>
class Vec3 : public Vec<T, 3>
{
public:

  inline Vec3();
  inline Vec3(const Vec3<T> &);
  inline Vec3(const Vec<T, 3> &);
  inline Vec3(T, T, T);
  inline Vec3(T);
  
  T & x, & y, & z;
  
  Vec3<T> & operator = (const Vec3<T> & in) { set(in); return *this; }
};

//------------------------------------------------------------------------------
// Functions

template<typename T, int D>
inline Vec<T, D> abs(const Vec<T, D> &);

#if defined(min) // XXX WIN32.
#undef min
#endif
#if defined(max)
#undef max
#endif
template<typename T, int D>
inline Vec<T, D> min(const Vec<T, D> &, const Vec<T, D> &);
template<typename T, int D>
inline Vec<T, D> max(const Vec<T, D> &, const Vec<T, D> &);
template<typename T, int D>
inline Vec<T, D> clamp(const Vec<T, D> &, const Vec<T, D> &, const Vec<T, D> &);

template <typename T, int D>
inline Vec<T, D> normalize(const Vec<T, D> &);
template <typename T, int D>
inline double dot(const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline double length(const Vec<T, D> &);
template<typename T, int D>
inline bool size_valid(const Vec<T, D> &);

template<typename T>
inline double aspect(const Vec2<T> &);
template <typename T>
inline Vec2<T> swap(const Vec2<T> &);

//------------------------------------------------------------------------------
// Operators

template <typename T, int D>
inline Vec<T, D> operator + (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline Vec<T, D> operator - (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline Vec<T, D> operator * (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline Vec<T, D> operator / (const Vec<T, D> &, const Vec<T, D> &);

template <typename T, int D>
inline Vec<T, D> operator + (const Vec<T, D> &, T);
template <typename T, int D>
inline Vec<T, D> operator - (const Vec<T, D> &, T);
template <typename T, int D>
inline Vec<T, D> operator * (const Vec<T, D> &, T);
template <typename T, int D>
inline Vec<T, D> operator / (const Vec<T, D> &, T);

template <typename T, int D>
inline Vec<T, D> operator - (const Vec<T, D> &);

//------------------------------------------------------------------------------
// Comparison

template <typename T, int D>
inline bool operator == (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline bool operator != (const Vec<T, D> &, const Vec<T, D> &);

template <typename T, int D>
inline bool operator <  (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline bool operator <= (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline bool operator >  (const Vec<T, D> &, const Vec<T, D> &);
template <typename T, int D>
inline bool operator >= (const Vec<T, D> &, const Vec<T, D> &);

//------------------------------------------------------------------------------
// Conversion

template <int D>
inline Vec<int, D> to_int(const Vec<double, D> &);
template <int D>
inline Vec<double, D> to_float(const Vec<int, D> &);

template <int D>
inline Vec<int, D> ceil(const Vec<double, D> &);
template <int D>
inline Vec<int, D> floor(const Vec<double, D> &);

template <typename T, int D>
inline bool serialize(List<Str> *, Vec<T, D> *);
template <typename T, int D>
inline List<Str> serialize(const Vec<T, D> &);

//------------------------------------------------------------------------------
// Debugging

template <typename T, int D>
inline Dbg & operator << (Dbg &, const Vec<T, D> &);

}
}

#include <djv_base/vec_.h>
#include <djv_base/vec2_.h>
#include <djv_base/vec3_.h>
#include <djv_base/vec_fnc_.h>

#endif

