/*------------------------------------------------------------------------------
 lib/djv_image/app.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/app.cxx"

#include "app.h"
#include <djv_image/gl_data.h>
#include <djv_image/gl_offscreen.h>
#include <djv_image/io.h>
#include <FL/Fl.H>

namespace djv {
namespace image {
namespace app
{

App::App(const Str & name, int & argc, char ** argv) throw (Error) :
  djv::App(name, argc, argv)
{
  DJV_DBG_OBJ2(_DBG, "App::App");
  
  Fl::visual(FL_DOUBLE | FL_RGB);
    
  // Command line.

  if (djv::app::EXIT_DEFAULT == exit())
    try { cmdln(); }
    catch (Error in)
    {
      error(str::Format("Command Line: %1").arg(in));
      exit(djv::app::EXIT_ERROR);
      //return;
    }
  //if (exit() != djv::app::EXIT_DEFAULT) return;
  
  // Initialize.
  
  gl::offscreen::init();
  io::init();
}

App::~App()
{
  DJV_DBG_OBJ2(_DBG, "App::~App");

  io::del();
  gl::offscreen::del();
}

void App::cmdln() throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "App::cmdln");

  List<Str> tmp;
  while (cmdln_list.size())
  {
    const Str arg = cmdln_list.pop_front();
    
    if (str::compare("-gl_fixed", arg))
    {
      gl::fixed(true);
    }
    else if (str::compare("-gl_texture_tile", arg))
    {
      Vec2i value;
      if (! vec::serialize(&cmdln_list, &value)) throw Error(arg);
      gl::data::texture_tile(value);
    }
    else if (str::compare("-gl_texture_dbg", arg))
    {
      gl::data::texture_dbg(true);
    }
    else if (str::compare("-gl_offscreen_tile", arg))
    {
      Vec2i value;
      if (! vec::serialize(&cmdln_list, &value)) throw Error(arg);
      gl::offscreen::tile(value);
    }
    else if (str::compare("-gl_offscreen_share", arg))
    {
      bool value;
      if (! str::serialize(&cmdln_list, &value)) throw Error(arg);
      gl::offscreen::context_share(value);
    }
    else
    {
      tmp += arg;
    }
  }
  
  cmdln_list = tmp;
}

namespace
{
const Str _help =
"OpenGL Options:\n"
"\n"
"-gl_fixed - Fixed function pipeline (OpenGL 1.2).\n"
"-gl_texture_tile (width) (height) - Texture tile size. Default: %1.\n"
"-gl_texture_dbg - Debug texture tiling.\n"
"-gl_offscreen_tile (width) (height) - Offscreen rendering tile size. "
"Default: %2.\n"
"-gl_offscreen_share (value) - Share offscreen rendering contexts. "
"Default: %3.\n"
"\n"
"%3";
}

Str App::help() const
{
  return str::Format(_help).
    arg(vec::serialize(gl::data::texture_tile())).
    arg(vec::serialize(gl::offscreen::tile())).
    arg(str::serialize(gl::offscreen::context_share())).
    arg(djv::App::help());
}

namespace
{
const Str _info =
"%1\n"
"\n"
"OpenGL:\n"
"\n"
"* fixed: %2\n"
"* texture tile: %3\n"
"* texture debug: %4\n"
"* offscreen tile: %5\n"
"* offscreen share: %6\n"
"\n"
"Image I/O:\n"
"\n"
"* load: %7\n"
"* save: %8";
}

Str App::info() const
{
  return str::Format(_info).
    arg(djv::App::info()).
    arg(str::serialize(gl::fixed())).
    arg(vec::serialize(gl::data::texture_tile())).
    arg(str::serialize(gl::data::texture_dbg())).
    arg(vec::serialize(gl::offscreen::tile())).
    arg(str::serialize(gl::offscreen::context_share())).
    arg(io::load()->name_list(), ", ").
    arg(io::save()->name_list(), ", ");
}

App * global()
{
  return static_cast<djv::image::App *>(djv::app::global());
}

}
}
}

