/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_fixed.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_data_fixed.cxx"

#include "_gl_data.h"
#include <djv_base/mtx.h>

namespace djv {
namespace image {
namespace gl {
namespace data
{

namespace
{
const Str
  error_namespace = "OpenGL";
}

image::Info color_table_info(GLenum target, const image::Info & in)
{
  image::Info out(in);
  GLint size = 0, format = gl::format(out.pixel, image::data::BGR == out.color),
    type = gl::type(out.pixel);
  do
  {
    if (GLEW_ARB_imaging)
    {
      glColorTable(target, GL_RGBA, math::to_pow2(out.size.x), format, type,
        0);
      glGetColorTableParameteriv(target, GL_COLOR_TABLE_WIDTH, &size);
    }
    else if (GL_SGI_color_table)
    {
      glColorTableSGI(target, GL_RGBA, math::to_pow2(out.size.x), format, type,
        0);
      glGetColorTableParameterivSGI(target, GL_COLOR_TABLE_WIDTH_SGI, &size);
    }
    if (! size) out.size.x = out.size.x >> 2;
  } while (! size && out.size.x);
  return out;
}

void Data::_Fixed::draw(const Draw & in, void * context_share) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Data::_Fixed::draw");

  // Profile.
  
  if (profile::LUT == in.profile.type &&
    vec::size_valid(in.profile.lut.size()))
  {
    GLint max = 0;
    DJV_DBG_GL(glGetIntegerv(GL_MAX_PIXEL_MAP_TABLE, &max));
    const int size = math::min(math::to_pow2(in.profile.lut.size().x),
      int(max));
    //DJV_DBG_PRN("pixel map max = " << max);
    DJV_DBG_PRN("profile lut = " << size);
    
    image::Data tmp(image::data::Info(Vec2i(size, 1), pixel::RGBA_F32));
    Draw draw;
    draw.scale = vec::to_float(tmp.info().size);
    data::copy(in.profile.lut, &tmp, draw, context_share);

    mem::Buffer<GLfloat> data(size * 4);
    const pixel::F32_T * in_p = reinterpret_cast<pixel::F32_T *>(tmp.data());
    GLfloat * out_p = data();
    for (int i = 0; i < size; ++i, in_p += 4, out_p += 1)
    {
      //DJV_DBG_PRN("lut[" << i << "] = " << in_p[0]);
      out_p[size * 0] = in_p[0];
      out_p[size * 1] = in_p[1];
      out_p[size * 2] = in_p[2];
      out_p[size * 3] = in_p[3];
    }

    DJV_DBG_GL(glPixelTransferi(GL_MAP_COLOR, GL_TRUE));
    DJV_DBG_GL(glPixelMapfv(GL_PIXEL_MAP_R_TO_R, size, data() + size * 0));
    DJV_DBG_GL(glPixelMapfv(GL_PIXEL_MAP_G_TO_G, size, data() + size * 1));
    DJV_DBG_GL(glPixelMapfv(GL_PIXEL_MAP_B_TO_B, size, data() + size * 2));
    DJV_DBG_GL(glPixelMapfv(GL_PIXEL_MAP_A_TO_A, size, data() + size * 3));
  }

  // Display.
  
  if ((GLEW_ARB_imaging || GL_SGI_color_table) &&
    vec::size_valid(in.display.lut.size()))
  {
    GLenum target = 0;
    if (GLEW_ARB_imaging) target = GL_PROXY_COLOR_TABLE;
    else if (GL_SGI_color_matrix) target = GL_PROXY_COLOR_TABLE_SGI;
    const image::Info info(color_table_info(target,
      image::Info(Vec2i(in.display.lut.size().x, 1), pixel::RGBA_U8)));
    DJV_DBG_PRN("display lut = " << info);
    
    image::Data tmp(info);
    Draw draw;
    draw.scale = vec::to_float(tmp.info().size);
    data::copy(in.display.lut, &tmp, draw, context_share);

    unpack(info);
    if (GLEW_ARB_imaging)
    {
      target = GL_COLOR_TABLE;
      DJV_DBG_GL(glColorTable(target, GL_RGBA, info.size.x, GL_RGBA,
        GL_UNSIGNED_BYTE, tmp.data()));
    }
    else if (GL_SGI_color_matrix)
    {
      target = GL_COLOR_TABLE_SGI;
      DJV_DBG_GL(glColorTableSGI(target, GL_RGBA, info.size.x, GL_RGBA,
        GL_UNSIGNED_BYTE, tmp.data()));
    }
    DJV_DBG_GL(glEnable(target));
  }
  
  if (GLEW_ARB_imaging || GL_SGI_color_matrix)
  {
    DJV_DBG_PRN("color matrix");

    DJV_DBG_GL(glMatrixMode(GL_COLOR));
    DJV_DBG_GL(glLoadMatrixf(color_mtx(in.display.color).e));
    DJV_DBG_GL(glMatrixMode(GL_MODELVIEW));
  }

  if ((GLEW_ARB_imaging || GL_SGI_color_table) &&
    (in.display.levels != Display().levels ||
      in.display.soft_clip != Display().soft_clip))
  {
    GLenum target = 0;
    if (GLEW_ARB_imaging)
      target = GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE;
    else if (GL_SGI_color_matrix)
      target = GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE_SGI;

    const image::Data lut = color_lut(in.display.levels, in.display.soft_clip);
    const image::Info info(color_table_info(target,
      image::Info(Vec2i(lut.size().x, 1), pixel::RGBA_U8)));
    DJV_DBG_PRN("color lut = " << info);

    image::Data tmp(info);
    Draw draw;
    draw.scale = vec::to_float(tmp.info().size);
    data::copy(lut, &tmp, draw, context_share);

    unpack(info);
    if (GLEW_ARB_imaging)
    {
      target = GL_POST_COLOR_MATRIX_COLOR_TABLE;
      DJV_DBG_GL(glColorTable(target, GL_RGBA, info.size.x, GL_RGBA,
        GL_UNSIGNED_BYTE, tmp.data()));
    }
    else if (GL_SGI_color_matrix)
    {
      target = GL_POST_COLOR_MATRIX_COLOR_TABLE_SGI;
      DJV_DBG_GL(glColorTableSGI(target, GL_RGBA, info.size.x, GL_RGBA,
        GL_UNSIGNED_BYTE, tmp.data()));
    }
    DJV_DBG_GL(glEnable(target));
  }
  
  if (in.channel > 0)
  {
    DJV_DBG_PRN("channel = " << in.channel);
    
    bool mask [] = { GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE };
    mask[int(in.channel) - 1] = GL_TRUE;
    DJV_DBG_PRN("mask = " << mask[0] << " " << mask[1] << " " << mask[2] <<
      " " << mask[3]);
    DJV_DBG_GL(glColorMask(mask[0], mask[1], mask[2], mask[3]));
  }
}

}
}
}
}

