/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_texture.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_data_texture.cxx"

#include "_gl_data.h"
#include <djv_base/mtx.h>

namespace djv {
namespace image {
namespace gl {
namespace data
{

namespace
{
const Str
  error_namespace = "OpenGL";

GLint gl_filter(FILTER in)
{
  switch (in)
  {
    case FILTER_NEAREST: return GL_NEAREST;
    case FILTER_LINEAR: return GL_LINEAR;
  }
  return 0;
}
}

Data::_Texture::_Texture() :
  _filter(FILTER(0)),
  _init(false), _texture(0)
{
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::_Texture");
}

Data::_Texture::~_Texture()
{
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::~_Texture");
  
  if (_texture)
  {
    DJV_DBG_PRN("delete texture");
    DJV_DBG_GL(glDeleteTextures(1, &_texture));
  }
}

namespace
{
inline void _quad(GLenum in, const Box2i & geom, const double * u = 0,
  const double * v = 0)
{
  glBegin(in);
  if (u && v) glTexCoord2d(u[0], v[0]);
  glVertex2i(geom.x, geom.y);
  if (u && v) glTexCoord2d(u[0], v[1]);
  glVertex2i(geom.x, geom.y + geom.h);
  if (u && v) glTexCoord2d(u[1], v[1]);
  glVertex2i(geom.x + geom.w, geom.y + geom.h);
  if (u && v) glTexCoord2d(u[1], v[0]);
  glVertex2i(geom.x + geom.w, geom.y);
  glEnd();
}
}

void Data::_Texture::draw(const image::Data & in, const Box2i & area,
  const Draw & draw, void * context_share) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::draw");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("scale = " << draw.scale);
  
  const bool texture_dbg = data::texture_dbg();
  
  if (draw.filter != _filter)
  {
    _init = false;
    _filter = draw.filter;
  }
  const image::data::Info & info = in.info();
  if (info != _info)
  {
    _init = false;
    _info = info;
  }
  if (texture_tile() != _texture_size)
  {
    _init = false;
    _texture_size = texture_tile();
  }
  DJV_DBG_PRN("texture size = " << _texture_size);
  
  Box2i viewport = area;
  if (texture_dbg) viewport = box::expand(viewport, Vec2i(-100));
  DJV_DBG_PRN("viewport = " << viewport);
  
  const int proxy_scale = image::data::proxy_scale(info.proxy);
  const int w = info.size.x * proxy_scale, h = info.size.y * proxy_scale;

  const Vec2f scale = vec::size_valid(draw.scale) ?
    (draw.scale / Vec2f(w, h)) : 1.0;
  DJV_DBG_PRN("scale = " << scale);
  
  const bool mirror_x = info.mirror_x ? (! draw.mirror_x) : draw.mirror_x,
    mirror_y = info.mirror_y ? (! draw.mirror_y) : draw.mirror_y;
  DJV_DBG_PRN("mirror = " << mirror_x << " " << mirror_y);
  
  // Initialize texture.
  
  GLenum target = GL_TEXTURE_2D;
  if (! gl::fixed())
  {
    if (GLEW_ARB_texture_rectangle) target = GL_TEXTURE_RECTANGLE_ARB;
    DJV_DBG_GL(glActiveTexture(GL_TEXTURE0));
  }

  if (! _init)
  {
    DJV_DBG_PRN("init");

    if (_texture)
    {
      DJV_DBG_PRN("delete texture");
      DJV_DBG_GL(glDeleteTextures(1, &_texture));
    }
    DJV_DBG_GL(glGenTextures(1, &_texture));
    DJV_DBG_GL(glBindTexture(target, _texture));
    DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP));
    DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP));
    DJV_DBG_GL(glTexParameteri(target,
      //GL_TEXTURE_MAG_FILTER, gl_filter(_filter)));
      GL_TEXTURE_MAG_FILTER, GL_NEAREST));
    DJV_DBG_GL(glTexParameteri(target,
      //GL_TEXTURE_MIN_FILTER, gl_filter(_filter)));
      GL_TEXTURE_MIN_FILTER, GL_NEAREST));
    DJV_DBG_GL(glTexImage2D(
      target, 0,
      ! gl::fixed() && // XXX
        (pixel::F16 == pixel::type(info.pixel) ||
         pixel::F32 == pixel::type(info.pixel)) ?
        GL_FLOAT_RGBA_NV : GL_RGBA,
      _texture_size.x, _texture_size.y, 0,
      gl::format(info.pixel, image::data::BGR == info.color),
      gl::type(info.pixel),
      0));

    _init = true;
  }
  
  // Draw.

  const Mtx3f mtx =
    mtx::rotate3f(draw.rotate) *
    mtx::scale3f(scale) *
    mtx::translate3f(draw.position);

  DJV_DBG_GL(glPushMatrix());
  glTranslated(draw.position.x, draw.position.y, 0);
  glScaled(scale.x, scale.y, 1);
  glRotated(draw.rotate, 0, 0, 1);

  if (_texture) { DJV_DBG_GL(glBindTexture(target, _texture)); }
  DJV_DBG_GL(glEnable(target));
  glColor3f(1, 1, 1);

  List<Box2i> dbg_visible, dbg_discard;
  for (int y = 0; y < info.size.y; y += _texture_size.y)
    for (int x = 0; x < info.size.x; x += _texture_size.x)
    {
      const Box2i tile(x, y,
        math::min(_texture_size.x, info.size.x - x),
        math::min(_texture_size.y, info.size.y - y));
      
      const Box2i tile_xform = tile * proxy_scale;
      
      const Vec2i pt [] = {
        tile_xform.position,
        tile_xform.position + Vec2i(0, tile_xform.h),
        tile_xform.position + tile_xform.size,
        tile_xform.position + Vec2i(tile_xform.w, 0) };
      Vec2f min, max;
      for (uint j = 0; j < 4; ++j)
      {
        const Vec2f _pt = mtx * vec::to_float(pt[j]);
        if (! j) min = max = _pt;
        else
        {
          min = vec::min(_pt, min);
          max = vec::max(_pt, max);
        }
      }
      Box2i bbox;
      bbox.position = vec::to_int(min);
      bbox.position1(vec::to_int(max));
        
      double u [] = { 0, 0 }, v [] = { 0, 0 };
      u[! mirror_x] = tile.w;
      v[! mirror_y] = tile.h;
      if (GL_TEXTURE_2D == target)
      {
        u[! mirror_x] /= double(_texture_size.x);
        v[! mirror_y] /= double(_texture_size.y);
      }

      if (! vec::size_valid(box::intersect(viewport, bbox).size))
      {
        if (texture_dbg) dbg_discard += tile_xform;
        continue;
      }
      if (texture_dbg) dbg_visible += tile_xform;
      DJV_DBG_PRN("bbox = " << bbox);

      unpack(info, Vec2i(
        mirror_x ? (info.size.x - tile.w - tile.x) : tile.x,
        mirror_y ? (info.size.y - tile.h - tile.y) : tile.y));
      DJV_DBG_GL(glTexSubImage2D(
        target, 0,
        0, 0, tile.w, tile.h,
        gl::format(info.pixel, image::data::BGR == info.color),
        gl::type(info.pixel),
        in.data()));

      _quad(GL_QUADS, tile_xform, u, v);
    }
  DJV_DBG_PRN("tiles = " << dbg_visible.size() + dbg_discard.size());
  DJV_DBG_PRN("visible = " << dbg_visible.size());
  DJV_DBG_PRN("discard = " << dbg_discard.size());
  
  DJV_DBG_GL(glDisable(target));
  
  if (texture_dbg)
  {
    glLineWidth(1);
    glColor3f(0, 0, 1);
    for (uint i = 0; i < dbg_discard.size(); ++i)
      _quad(GL_LINE_LOOP, dbg_discard[i]);
    glColor3f(1, 1, 0);
    for (uint i = 0; i < dbg_visible.size(); ++i)
      _quad(GL_LINE_LOOP, dbg_visible[i]);
  }
  
  DJV_DBG_GL(glPopMatrix());
  
  if (texture_dbg)
  {
    glLineWidth(2);
    glColor3f(1, 0, 0);
    _quad(GL_LINE_LOOP, viewport);
  }
}

}
}
}
}

