/*------------------------------------------------------------------------------
 lib/djv_image/gl_offscreen_glx.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_offscreen_glx.cxx"

#include "gl_offscreen.h"
#include <djv_base/vec.h>
#include <FL/x.H>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xmu/StdCmap.h>

namespace djv {
namespace image {
namespace gl {
namespace offscreen
{

namespace
{
const Str
  error_glew = "Initialize: %1",
  error_glx = "Not Found",
  error_glx_fbconfig = "Frame Buffer Configuration",
  error_glx_pbuffer = "PBuffer",
  error_glx_context = "Context",
  error_glx_bind = "Context Bind",
  error_x11_display = "Open Display",
  error_x11_visual = "Visual",
  error_x11_colormap = "Colormap",
  error_x11_window = "Window";

class _Pbuffer
{
public:

  _Pbuffer(const Vec2i & size, void * context_share) throw (Error);
  ~_Pbuffer();
  
  void current() throw (Error);
  
  const Vec2i & size() const { return _size; }
  void * context_share() const { return _context_share; }
  
private:

  Vec2i _size;
  GLXContext _context;
  void * _context_share;
  GLXPbuffer _pbuffer;
};

class _Pbuffer_Factory
{
public:

  _Pbuffer_Factory() throw (Error);
  ~_Pbuffer_Factory();
  
  void add(_Pbuffer *);
  
  _Pbuffer * get(const Vec2i & size, void * context_share) throw (Error);
  
  GLXFBConfig config() const { return _config; }
  
private:

  List<_Pbuffer *> _list;
  Window _window;
  GLXContext _context;
  GLXFBConfig _config;
};

std::auto_ptr<_Pbuffer_Factory> _factory;

_Pbuffer::_Pbuffer(const Vec2i & size, void * context_share) throw (Error) :
  _context(0), _context_share(context_share), _pbuffer(0)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::_Pbuffer");
  DJV_DBG_PRN("size = " << size);
  
  const GLXFBConfig config = _factory->config();
  if (! fl_display || ! config) return;
  
  _context = glXCreateNewContext(fl_display, config, GLX_RGBA_TYPE,
    (GLXContext)_context_share, True);
  if (! _context) throw Error("GLX", error_glx_context);

  int attrib [] = {
    GLX_PBUFFER_WIDTH, size.x,
    GLX_PBUFFER_HEIGHT, size.y,
    GLX_PRESERVED_CONTENTS, True,
    //GLX_LARGEST_PBUFFER, True,
    0 };
  _pbuffer = glXCreatePbuffer(fl_display, config, attrib);
  if (! _pbuffer) throw Error("GLX", error_glx_pbuffer);
  
  GLuint w = 0, h = 0;
  glXQueryDrawable(fl_display, _pbuffer, GLX_WIDTH_SGIX, &w);
  glXQueryDrawable(fl_display, _pbuffer, GLX_HEIGHT_SGIX, &h);
  _size = Vec2i(w, h);
  DJV_DBG_PRN("real size = " << _size);
}

_Pbuffer::~_Pbuffer()
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::~_Pbuffer");

  if (_pbuffer)
  {
    DJV_DBG_PRN("delete pbuffer");
    glXDestroyPbuffer(fl_display, _pbuffer);
  }
  if (_context)
  {
    DJV_DBG_PRN("delete context");
    glXDestroyContext(fl_display, _context);
  }
}

void _Pbuffer::current() throw (Error)
{
  if (! glXMakeCurrent(fl_display, _pbuffer, _context))
    throw Error("glx", error_glx_bind);
}

namespace
{
struct Fb_Config
{
  GLXFBConfig p;
  int id;
  int visual, visual_type;
  int caveat;
  int r, g, b, a;
};

inline int weight_visual_type(int in)
{
  switch (in)
  {
    case GLX_NONE: return 7;
    case GLX_TRUE_COLOR: return 6;
    case GLX_DIRECT_COLOR: return 5;
    case GLX_PSEUDO_COLOR: return 4;
    case GLX_STATIC_COLOR: return 3;
    case GLX_GRAY_SCALE: return 2;
    case GLX_STATIC_GRAY: return 1;
  }
  return 0;
}

inline int weight_caveat(int in)
{
  switch (in)
  {
    case GLX_NONE: return 3;
    case GLX_SLOW_CONFIG: return 2;
    case GLX_NON_CONFORMANT_CONFIG: return 1;
  }
  return 0;
}

Str label_visual(int in)
{
  char out [str::cstr_size];
  SNPRINTF(out, str::cstr_size, "%x", in);
  return out;
}

Str label_visual_type(int in)
{
  switch (in)
  {
    case GLX_TRUE_COLOR: return "true color";
    case GLX_DIRECT_COLOR: return "direct color";
    case GLX_PSEUDO_COLOR: return "pseudo color";
    case GLX_STATIC_COLOR: return "static color";
    case GLX_GRAY_SCALE: return "gray scale";
    case GLX_STATIC_GRAY: return "static gray";
  }
  return "none";
}

Str label_caveat(int in)
{
  switch (in)
  {
    case GLX_SLOW_CONFIG: return "slow";
    case GLX_NON_CONFORMANT_CONFIG: return "nonconformant";
  }
  return "none";
}

inline bool compare(const Fb_Config & a, const Fb_Config & b)
{
  return
    (b.r + b.g + b.b + b.a + weight_visual_type(b.visual_type) +
      weight_caveat(b.caveat)) <
    (a.r + a.g + a.b + a.a + weight_visual_type(a.visual_type) +
      weight_caveat(a.caveat));
}

GLXFBConfig fb_config(Display * display, int screen) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "fb_config");
  
  struct Glx
  {
    Glx(Display * display, int screen) throw (Error)
    {
      int attrib [] = {
        GLX_RENDER_TYPE, GLX_RGBA_BIT,
        GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT_SGIX,
        GLX_DOUBLEBUFFER, False,
        GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_ALPHA_SIZE, 0,
        None };
      list = glXChooseFBConfigSGIX(display, screen, attrib, &size);
    }
    
    ~Glx() { if (list) XFree(list); }
    
    GLXFBConfig * list;
    int size;

  } glx(display, screen);

  List<Fb_Config> list;
  for (int i = 0; i < glx.size; ++i)
  {
    Fb_Config tmp;
    tmp.p = glx.list[i];
    glXGetFBConfigAttrib(display, glx.list[i], GLX_FBCONFIG_ID, &tmp.id);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_VISUAL_ID, &tmp.visual);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_X_VISUAL_TYPE,
      &tmp.visual_type);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_CONFIG_CAVEAT,
      &tmp.caveat);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_RED_SIZE, &tmp.r);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_GREEN_SIZE, &tmp.g);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_BLUE_SIZE, &tmp.b);
    glXGetFBConfigAttrib(display, glx.list[i], GLX_ALPHA_SIZE, &tmp.a);
    
    if (gl::fixed())
    {
      if (tmp.r > 8 || tmp.g > 8  || tmp.b > 8 || tmp.a > 8) continue;
    }
    
    list += tmp;
  }
  if (! list.size()) throw Error("GLX", error_glx_fbconfig);
  
  std::sort(list.begin(), list.end(), compare);

#ifdef DJV_DBG
  for (int i = 0; i < list.size(); ++i)
    DJV_DBG_PRN("id " << list[i].id <<
      " visual = " << label_visual(list[i].visual) <<
      " visual type = " << label_visual_type(list[i].visual_type) <<
      " caveat " << label_caveat(list[i].caveat) <<
      " depth = " << list[i].r << " " << list[i].g << " " <<
      list[i].b << " " << list[i].a);
#endif
    
  return list[0].p;
}

}

_Pbuffer_Factory::_Pbuffer_Factory() throw (Error) :
  _window(0), _context(0), _config(0)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::_Pbuffer_Factory");
  
  fl_open_display();
  
  if (! glXQueryExtension(fl_display, 0, 0)) throw Error("GLX", error_glx);

  // XXX Create a dummy window and rendering context for glewInit.
  // According to the docs, glewInit can be called just once per-process?
  
  XSetWindowAttributes win_attrib;
  win_attrib.colormap = fl_colormap;
  win_attrib.border_pixel = 0;
  _window = XCreateWindow(fl_display, RootWindow(fl_display, fl_screen),
    0, 0, 1, 1, 0, fl_visual->depth, InputOutput, fl_visual->visual,
    CWColormap | CWBorderPixel, &win_attrib);
  if (! _window) throw Error("X11", error_x11_window);

  _context = glXCreateContext(fl_display, fl_visual, 0, True);
  if (! _context) throw Error("GLX", error_glx_context);
  
  if (! glXMakeCurrent(fl_display, _window, _context))
    throw Error("GLX", error_glx_bind);
  
  // Initialize GLEW.
  
  GLint error = glewInit();
  if (error != GLEW_OK)
    throw Error("GLEW", str::Format(error_glew).
      arg((char *)glewGetErrorString(error)));
  
  // XXX
  
  if (! gl::fixed()) gl::fixed(! GLEW_ARB_fragment_shader);

  // Frame buffer configuration.
  
  _config = fb_config(fl_display, fl_screen);
}

_Pbuffer_Factory::~_Pbuffer_Factory()
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::~_Pbuffer_Factory");

  for (uint i = 0; i < _list.size(); ++i) delete _list[i];
  if (_context) glXDestroyContext(fl_display, _context);
  if (_window) XDestroyWindow(fl_display, _window);
}

void _Pbuffer_Factory::add(_Pbuffer * in)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::add");

  if (0)
  {
    delete in;
  }
  else
  {
    _list += in;
    if (int(_list.size()) > 10)
      delete _list.pop_front();
  }
}

_Pbuffer * _Pbuffer_Factory::get(const Vec2i & size, void * context_share)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::get");
  DJV_DBG_PRN("size = " << size);
  
  _Pbuffer * out = 0;

  for (List<_Pbuffer *>::iterator i = _list.begin(); i != _list.end(); ++i)
    if (size <= (*i)->size() && (*i)->context_share() == context_share)
    {
      DJV_DBG_PRN("cached");
      out = *i;
      _list.erase(i);
      break;
    }
  
  if (! out) out = new _Pbuffer(size, context_share);
  return out;
}

}

class Pbuffer::Context_Restore
{
public:

  Context_Restore() : _context(0), _drawable(0)
  {
    _context = glXGetCurrentContext();
    _drawable = glXGetCurrentDrawable();
  }
  
  ~Context_Restore()
  {
    glXMakeCurrent(fl_display, _drawable, _context);
  }
  
private:

  GLXContext _context;
  GLXDrawable _drawable;
};

Pbuffer::Pbuffer(const Vec2i & size, void * context_share) throw (Error) :
  _p(0),
  _context_restore(new Context_Restore)
{
  DJV_DBG_OBJ2(_DBG, "Pbuffer::Pbuffer");
  
  _p = _factory->get(offscreen::tile(),
    offscreen::context_share() ? context_share : 0);
  _p->current();
}

Pbuffer::~Pbuffer()
{
  DJV_DBG_OBJ2(_DBG, "Pbuffer::~Pbuffer");
  
  if (_p) _factory->add(_p);
}

const Vec2i & Pbuffer::size() const { return _p->size(); }

void init() throw (Error)
{ _factory = std::auto_ptr<_Pbuffer_Factory>(new _Pbuffer_Factory); }
void del() { _factory.reset(); }

}
}
}
}

