/*------------------------------------------------------------------------------
 lib/djv_image/gl_offscreen_wgl.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_offscreen_wgl.cxx"

#include "gl_offscreen.h"
#include <djv_base/vec.h>
#include <djv_image/app.h>
#include <FL/Fl.H>

namespace djv {
namespace image {
namespace gl {
namespace offscreen
{

namespace
{
const Str
  error_glew = "Initialize: %1",
  error_glew_pbuffer = "PBuffer",
  error_wgl_pbuffer = "PBuffer: %1",
  error_wgl_pbuffer_device = "PBuffer Device: %1",
  error_wgl_context = "Context: %1",
  error_wgl_bind = "Context Bind: %1",
  error_wgl_choose_pixel = "Choose Pixel: %1",
  error_wgl_pixel = "Pixel: %1",
  error_win32_module = "Module Instance: %1",
  error_win32_register = "Register Class: %1",
  error_win32_window = "Window: %1",
  error_win32_device = "Device: %1",
  error_win32_choose_pixel = "Choose Pixel: %1",
  error_win32_pixel = "Pixel: %1";

class _Pbuffer
{
public:

  _Pbuffer(const Vec2i & size, void * context_share) throw (Error);
  ~_Pbuffer();
  
  void current() throw (Error);
  
  const Vec2i & size() const { return _size; }
  void * context_share() const { return _context_share; } 

private:

  Vec2i _size;
  HDC _device;
  HGLRC _context;
  void * _context_share;
  HPBUFFERARB _pbuffer;
};

class _Pbuffer_Factory
{
public:

  _Pbuffer_Factory() throw (Error);
  ~_Pbuffer_Factory();
  
  void add(_Pbuffer *);
  _Pbuffer * get(const Vec2i & size, void * context_share) throw (Error);
  
  int format() const { return _format; }
  HDC device() const { return _device; }
  
private:

  List<_Pbuffer *> _list;
  int _format;
  HWND _id;
  HDC _device;
  HGLRC _context;
};

std::auto_ptr<_Pbuffer_Factory> _factory;

_Pbuffer::_Pbuffer(const Vec2i & size, void * context_share) throw (Error) :
  _context(0), _context_share(context_share), _pbuffer(0)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::_Pbuffer");
  DJV_DBG_PRN("size = " << size);
  
  int attrib [] = {
    WGL_PBUFFER_LARGEST_ARB, 1,
    0 };
  _pbuffer = wglCreatePbufferARB(_factory->device(), _factory->format(),
    size.x, size.y, attrib);
  if (! _pbuffer)
    throw Error("WGL", str::Format(error_wgl_pbuffer).arg(int(GetLastError())));
  wglQueryPbufferARB(_pbuffer, WGL_PBUFFER_WIDTH_ARB, &_size.x);
  wglQueryPbufferARB(_pbuffer, WGL_PBUFFER_HEIGHT_ARB, &_size.y);
  DJV_DBG_PRN("real size = " << _size);
  
  _device = wglGetPbufferDCARB(_pbuffer);
  if (! _device)
    throw Error("WGL", str::Format(error_wgl_pbuffer_device).
      arg(int(GetLastError())));

  _context = wglCreateContext(_device);
  if (! _context)
    throw Error("WGL", str::Format(error_wgl_context).arg(int(GetLastError())));
  wglShareLists((HGLRC)context_share, _context);
}

_Pbuffer::~_Pbuffer()
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::~_Pbuffer");
  
  if (_context)
  {
    wglDeleteContext(_context);
    _context = 0;
  }
  if (_pbuffer)
  {
    if (_device)
    {
      wglReleasePbufferDCARB(_pbuffer, _device);
      _device = 0;
    }
    wglDestroyPbufferARB(_pbuffer);
    _pbuffer = 0;
  }
}
    
void _Pbuffer::current() throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::current");

  if (! wglMakeCurrent(_device, _context))
    throw Error("WGL", str::Format(error_wgl_bind).arg(int(GetLastError())));
}

_Pbuffer_Factory::_Pbuffer_Factory() throw (Error) :
  _format(0), _id(0), _device(0), _context(0)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::_Pbuffer_Factory");

  // XXX Create a dummy window and rendering context for glewInit.
  // According to the docs, glewInit can be called just once per-process?
  
  HINSTANCE hinstance = GetModuleHandle(0);
  if (! hinstance)
    throw Error("WIN32", str::Format(error_win32_module).
      arg(int(GetLastError())));
  
  static const char name [] = "djv";
  WNDCLASS wc;
  if (! GetClassInfo(hinstance, name, &wc))
  {
    wc.style = CS_OWNDC;
    //wc.lpfnWndProc = (WNDPROC)MainWndProc;
    wc.lpfnWndProc = DefWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hinstance;
    wc.hIcon = LoadIcon(0, IDI_APPLICATION);
    wc.hCursor = LoadCursor(0, IDC_ARROW);
    wc.hbrBackground = 0;
    wc.lpszMenuName = 0;
    wc.lpszClassName = name;
    if (! RegisterClass(&wc))
      throw Error("WIN32", str::Format(error_win32_register).
        arg(int(GetLastError())));
  }
  
  _id = CreateWindow(name, 0, 0, 0, 0, 0, 0, 0, 0, hinstance, 0);
  if (! _id)
    throw Error("WIN32", str::Format(error_win32_window).
      arg(int(GetLastError())));
  
  _device = GetDC(_id);
  if (! _device)
    throw Error("WIN32", str::Format(error_win32_device).
      arg(int(GetLastError())));

  PIXELFORMATDESCRIPTOR pixel_format = { 
    sizeof(PIXELFORMATDESCRIPTOR),
    1,
    PFD_SUPPORT_OPENGL,
    PFD_TYPE_RGBA,
    24,
    0, 0, 0, 0, 0, 0,
    0,
    0,
    0,
    0, 0, 0, 0,
    0,
    0,
    0,
    PFD_MAIN_PLANE,
    0,
    0, 0, 0 };
  int pixel_format_id = ChoosePixelFormat(_device, &pixel_format);
  if (! pixel_format_id)
    throw Error("WIN32", str::Format(error_win32_choose_pixel).
      arg(int(GetLastError())));
  
  if (! SetPixelFormat(_device, pixel_format_id, &pixel_format))
    throw Error("WIN32", str::Format(error_win32_pixel).
      arg(int(GetLastError())));
  
  _context = wglCreateContext(_device);
  if (! _context)
    throw Error("WGL", str::Format(error_wgl_context).arg(int(GetLastError())));
  
  if (! wglMakeCurrent(_device, _context))
    throw Error("WGL", str::Format(error_wgl_bind).arg(int(GetLastError())));
  
  GLenum err = glewInit();
  if (err != GLEW_OK)
    throw Error("GLEW", str::Format(error_glew).
      arg((char *)glewGetErrorString(err)));
  
  if (! WGLEW_ARB_pbuffer)
    throw Error("GLEW", error_glew_pbuffer);
  
  // Pbuffer configuration.
  
  int fb_attrib [] = {
    WGL_SUPPORT_OPENGL_ARB, 1,
    WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
    WGL_DRAW_TO_PBUFFER_ARB, 1,
    WGL_DOUBLE_BUFFER_ARB, 0,
    WGL_RED_BITS_ARB, 1,
    WGL_GREEN_BITS_ARB, 1,
    WGL_BLUE_BITS_ARB, 1,
    WGL_ALPHA_BITS_ARB, 0,
    WGL_DEPTH_BITS_ARB, 0,
    WGL_STENCIL_BITS_ARB, 0,
    0 };
  uint format_size = 0;
  if (! wglChoosePixelFormatARB(_device, fb_attrib, 0, 1,
    &_format, &format_size))
    throw Error("WGL", str::Format(error_wgl_choose_pixel).
      arg(int(GetLastError())));
  if (! format_size)
    throw Error("WGL", str::Format(error_wgl_pixel).arg(int(GetLastError())));
}

_Pbuffer_Factory::~_Pbuffer_Factory()
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::~_Pbuffer_Factory");
  
  for (uint i = 0; i < _list.size(); ++i) delete _list[i];
  if (_context) wglDeleteContext(_context);
  if (_device) ReleaseDC(_id, _device);
}

void _Pbuffer_Factory::add(_Pbuffer * in)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::add");

  _list += in;
  if (int(_list.size()) > 10)
    delete _list.pop_front();
}

_Pbuffer * _Pbuffer_Factory::get(const Vec2i & size, void * context_share)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::get");
  DJV_DBG_PRN("size = " << size);
  
  _Pbuffer * out = 0;
  
  for (List<_Pbuffer *>::iterator i = _list.begin(); i != _list.end(); ++i)
    if (size <= (*i)->size() && (*i)->context_share() == context_share)
    {
      DJV_DBG_PRN("cached");
      out = *i;
      _list.erase(i);
      break;
    }

  if (! out) out = new _Pbuffer(size, context_share);
  out->current();
  return out;
}

}

class Pbuffer::Context_Restore
{
public:

  Context_Restore() : _context(0), _device(0)
  {
    _context = wglGetCurrentContext();
    _device = wglGetCurrentDC();
  }
  
  ~Context_Restore()
  {
    wglMakeCurrent(_device, _context);
  }
  
private:

  HGLRC _context;
  HDC _device;
};

Pbuffer::Pbuffer(const Vec2i & size, void * context_share) throw (Error) :
  _p(0),
  _context_restore(new Context_Restore)
{
  DJV_DBG_OBJ2(_DBG, "Pbuffer::Pbuffer");

  _p = _factory->get(offscreen::tile(),
    offscreen::context_share() ? context_share : 0);
}

Pbuffer::~Pbuffer()
{
  DJV_DBG_OBJ2(_DBG, "Pbuffer::~Pbuffer");

  if (_p) _factory->add(_p);
}

const Vec2i & Pbuffer::size() const { return _p->size(); }

void init() throw (Error)
{ _factory = std::auto_ptr<_Pbuffer_Factory>(new _Pbuffer_Factory); }
void del() { _factory.reset(); }

}
}
}
}

