/*------------------------------------------------------------------------------
 lib/djv_image/io.h
 
 Image I/O
 ---------
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_IO_H
#define DJV_IMAGE_IO_H

#include <djv_base/file.h>
#include <djv_base/io.h>
#include <djv_base/plugin.h>
#include <djv_image/image.h>
#include <djv_image/time.h>

namespace djv {
namespace image {
namespace io
{

//------------------------------------------------------------------------------
// Plugin Base

class Plugin : public plugin::Plugin
{
public:
  
  // List of supported file extensions.
  
  virtual const List<Str> & extension_list() const
  { return _null_extension_list; }
  
  // Command-line help string.
  
  virtual Str help() const { return Str(); }
  
  // Plugin options.
  
  virtual bool option(const Str &, List<Str> *) { return false; }
  virtual List<Str> option(const Str &) const { return List<Str>(); }
  virtual const List<Str> & option_list() const { return _null_option_list; }
  
private:

  List<Str> _null_extension_list, _null_option_list;
};

//------------------------------------------------------------------------------
// Load Plugin Base

class Load : public Plugin
{
public:
  
  // Open an image file/sequence and return it's information.
  
  virtual void open(const File &, Info *, Tag *, Time *)
    throw (Error) = 0;
  virtual void open(const File &, List<Info> *, Tag *, Time *)
    throw (Error) = 0;
  
  // Load images.
  
  virtual const Image * get(int frame = -1, int layer = 0, int proxy = 0)
    throw (Error) = 0;

protected:

  // Utility class for frame-based file formats.

  struct Cache
  {
    Cache() {}
    Cache(const File & file, int frame, int layer, int proxy) :
      file(file), frame(frame), layer(layer), proxy(proxy) {}

    File file;
    int frame, layer, proxy;

    bool operator == (const Cache & in) const
    {
      return in.file == file && in.frame == frame && in.layer == layer &&
        in.proxy == proxy;
    }
  };
};

//------------------------------------------------------------------------------
// Save Plugin Base

class Save : public Plugin
{
public:

  // Open an image file/sequence for writing.
  
  virtual void open(const File &, const Info &, const Tag &, const Time &)
    throw (Error) = 0;
  
  // Close.
  
  virtual void close() throw (Error) {}
  
  // Save images.
  
  virtual void set(const Image &, int frame = -1) throw (Error) = 0;
};

//------------------------------------------------------------------------------
// Load Plugin Factory

class Load_Factory : public plugin::Factory
{
public:

  Load_Factory();
  ~Load_Factory();
  
  // Test if a file is an image.
  
  bool valid(const File &, const Str & plugin = Str());

  // Find an image load plugin for the given file.
  
  Load * get(const File &, Info * = 0, Tag * = 0, Time * = 0,
    const Str & plugin = Str()) throw (Error);
  Load * get(const File &, List<Info> *, Tag * = 0, Time * = 0,
    const Str & plugin = Str()) throw (Error);
  
  // Command-line help string for all plugins.
  
  Str help() const;

  // Set a plugin option.
  
  bool option(const Str & plugin, const Str &, List<Str> *);
  
  // List of plugins.
  
  List<Load *> list() const;

private:

  virtual Str plugin_entry() const { return "djv_image_load"; }
};

//------------------------------------------------------------------------------
// Save Plugin Factory

class Save_Factory : public plugin::Factory
{
public:

  Save_Factory();
  ~Save_Factory();
  
  // Test if a file is an image.
  
  bool valid(const File &, const Str & plugin = Str());

  // Find an image save plugin for the given file.
  
  Save * get(const File &, const Info &, const Tag & tag = Tag(),
    const Time & = Time(), const Str & plugin = Str()) throw (Error);

  // Command-line help string for all plugins.
  
  Str help() const;

  // Set a plugin option.

  bool option(const Str & plugin, const Str &, List<Str> *);
  
  // List of plugins.
  
  List<Save *> list() const;

private:

  virtual Str plugin_entry() const { return "djv_image_save"; }
};

//------------------------------------------------------------------------------
// Labels

static const Str
  error_namespace = "Image I/O",
  error_unrecognized = "Unrecognized File",
  error_unsupported = "Unsupported File";
using djv::io::error_open;
using djv::io::error_read;
using djv::io::error_write;

//------------------------------------------------------------------------------
// Global

bool valid(const File &, const Str & plugin = Str());

Load * load(const File &, Info * = 0, Tag * = 0, Time * = 0,
  const Str & plugin = Str())throw (Error);
Load * load(const File &, List<Info> *, Tag * = 0, Time * = 0,
  const Str & plugin = Str()) throw (Error);
Save * save(const File &, const Info &, const Tag &, const Time &,
  const Str & plugin = Str()) throw (Error);

Load_Factory * load();
Save_Factory * save();

void init();
void del();

}
}
}

#endif

