/*------------------------------------------------------------------------------
 lib/djv_image/io_cineon.h

 Kodak Cineon Image I/O
 ----------------------

 References:
 * Kodak, "4.5 DRAFT - Image File Format Proposal for Digital Pictures"

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_IO_CINEON_H
#define DJV_IMAGE_IO_CINEON_H

#include <djv_image/io_dpx.h>

namespace djv {
namespace image {
namespace io {
namespace cineon
{

//------------------------------------------------------------------------------
// Plugin Information

static const Str name = "Cineon";

static const List<Str> extension_list = List<Str>() << ".cin";

//------------------------------------------------------------------------------
// Image Tags

enum TAG
{
  TAG_SOURCE_OFFSET, TAG_SOURCE_FILE, TAG_SOURCE_TIME, TAG_SOURCE_INPUT_DEVICE,
  TAG_SOURCE_INPUT_MODEL, TAG_SOURCE_INPUT_SERIAL, TAG_SOURCE_INPUT_PITCH,
  TAG_SOURCE_GAMMA,
  
  TAG_FILM_FORMAT, TAG_FILM_FRAME, TAG_FILM_FRAME_RATE, TAG_FILM_FRAME_ID,
  TAG_FILM_SLATE
};

//------------------------------------------------------------------------------
// Load Plugin

class Load : public io::Load
{
public:

  Load();
  
  Plugin * copy() const;

  Str name() const { return cineon::name; }
  const List<Str> & extension_list() const { return cineon::extension_list; }
  Str help() const;

  struct Options
  {
    Options() : color(dpx::COLOR_AUTO) {}
    dpx::COLOR color;
    dpx::Unprint print;
  };
  enum OPTION { OPTION_COLOR, OPTION_PRINT };
  bool option(const Str &, List<Str> *);
  List<Str> option(const Str &) const;
  const List<Str> & option_list() const;
  
  void open(const File &, Info *, Tag *, Time *) throw (Error);
  void open(const File & in, List<Info> * info, Tag * tag, Time * time)
    throw (Error)
  { io::Load::open(in, info, tag, time); }
  const Image * get(int frame, int layer, int proxy) throw (Error);

private:

  void _open(const File &, Info *, Tag *, time::Speed * = 0) throw (Error);

  Options _options;
  bool _print;
  std::auto_ptr<Data> _print_lut;
  File _file;
  djv::Io _io;
  Image _image;
  Data _tmp;
  Cache _cache;
};

//------------------------------------------------------------------------------
// Save Plugin

class Save : public io::Save {
public:
  
  Plugin * copy() const;

  Str name() const { return cineon::name; }
  const List<Str> & extension_list() const { return cineon::extension_list; }
  Str help() const;
  
  struct Options
  {
    Options() : color(dpx::COLOR_PRINT) {}
    dpx::COLOR color;
    dpx::Print print;
  };
  enum OPTION { OPTION_COLOR, OPTION_PRINT };
  bool option(const Str &, List<Str> *);
  List<Str> option(const Str &) const;
  const List<Str> & option_list() const;
  
  void open(const File &, const Info &, const Tag &, const Time &)
    throw (Error);
  void set(const Image &, int frame) throw (Error);

private:

  void _open(const File &, const Tag &) throw (Error);

  Options _options;
  File _file;
  djv::Io _io;
  Info _info;
  Image _image;
};

//------------------------------------------------------------------------------
// Functions

void header_load(djv::Io *, Info *, Tag *, time::Speed *, bool * print = 0)
  throw (Error);
void header_save(djv::Io *, const Str & file, const Info &, const Tag &,
  dpx::COLOR) throw (Error);
void header_save_update(djv::Io *) throw (Error);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_tag();
const List<Str> & label_load_option();
const List<Str> & label_save_option();

}
}
}
}

#endif

