/*------------------------------------------------------------------------------
 lib/djv_image/io_sgi_data.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_sgi_data.cxx"

#include "io_sgi.h"

namespace djv {
namespace image {
namespace io {
namespace _sgi
{

namespace
{

void _rle_unpack_u8(djv::Io * io, void * out, int size) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_rle_unpack_u8");
  DJV_DBG_PRN("size = " << size);
  
  uint8_t * out_p = reinterpret_cast<uint8_t *>(out);
  const uint8_t * const end = out_p + size;
  
  while (out_p < end)
  {
    // Get RLE information.

    uint8_t tmp;
    io->get_u8(&tmp);
    const int count = tmp & 0x7f;
    DJV_DBG_PRN("count = " << count);
    if (! count || count > (end - out_p)) break;
    const bool run = ! (tmp & 0x80);
    const int length = run ? 1 : count;
    DJV_DBG_PRN("  run = " << run);
    DJV_DBG_PRN("  length = " << length);

    // Unpack pixels.
    
    io->get(out_p, length);
    if (run)
      for (int j = 1; j < count; ++j)
        out_p[j] = out_p[0];

    out_p += count;
  }
  
  DJV_DBG_PRN("out = " << int(out_p - reinterpret_cast<uint8_t *>(out)));
}

void _rle_unpack_u16(djv::Io * io, void * out, int size) throw (Error)
{
  uint16_t * out_p = reinterpret_cast<uint16_t *>(out);
  const uint16_t * const end = out_p + size;

  while (out_p < end)
  {
    uint16_t tmp;
    io->get_u16(&tmp);
    const int count = tmp & 0x7f;
    if (! count || count > (end - out_p)) break;
    const bool run = ! (tmp & 0x80);
    const int length = run ? 1 : count;

    io->get(out_p, length, 2);
    if (run)
      for (int j = 1; j < count; ++j)
        out_p[j] = out_p[0];

    out_p += count;
  }
}

template<typename T>
size_t _rle_pack(const void * in, void * out, int size)
{
  DJV_DBG_OBJ2(_DBG, "_rle_pack");
  DJV_DBG_PRN("size = " << size);
  
  const T * in_p = reinterpret_cast<const T *>(in);
  T * out_p = reinterpret_cast<T *>(out);
  const T * const end = in_p + size;
  
  while (in_p < end)
  {
    // Find pixel runs.
    
    const int min = 3;
    const int max = math::min(0x7f, int(end - in_p));
    int count = 1, match = 1;
    for (; count < max; ++count)
    {
      if (in_p[count] == in_p[count - 1])
      {
        ++match;
        if (min == match && count >= min)
        {
          count -= min - 1;
          match = 1;
          break;
        }
      }
      else
      {
        if (match >= min) break;
        match = 1;
      }
    }
    const bool run = match > min;
    const int length = run ? 1 : count;
    DJV_DBG_PRN("count = " << count);
    DJV_DBG_PRN("  run = " << run);
    DJV_DBG_PRN("  length = " << length);
    
    // Set RLE information.
    
    out_p[0] = (count & 0x7f) | ((! run) << 7);
    
    // Pack pixels.
    
    for (int i = 0; i < length; ++i)
      out_p[1 + i] = in_p[i];
#   ifndef DJV_MSB
    mem::endian(out_p, 1 + length, sizeof(T));
#   endif
    
    out_p += 1 + length;
    in_p += count;
  }
  
  // Cap the end.
  
  *out_p++ = 0;
  
  const size_t r = (out_p - reinterpret_cast<T *>(out)) * sizeof(T);
  DJV_DBG_PRN("r = " << r);
  return r;
}

}

void rle_unpack(djv::Io * io, void * out, int size, int bytes) throw (Error)
{
  switch (bytes)
  {
    case 1: _rle_unpack_u8(io, out, size); break;
    case 2: _rle_unpack_u16(io, out, size); break;
  }
}

size_t rle_pack(const void * in, void * out, int size, int bytes)
{
  switch (bytes)
  {
    case 1: return _rle_pack<uint8_t>(in, out, size);
    case 2: return _rle_pack<uint16_t>(in, out, size);
  }
  
  return 0;
}

}
}
}
}
