/*------------------------------------------------------------------------------
 lib/djv_image/io_sgi_save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_sgi_save.cxx"

#include "io_sgi.h"

namespace djv {
namespace image {
namespace io {
namespace _sgi
{

Save::Save()
{
  _io.endian(mem::endian() != mem::MSB);
}

Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* sgi compress (value) - Compression: %1. Default: %2.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_compress(), ", ").
    arg(str::serialize(_options.compress, label_compress()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, label_compress());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, label_compress());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &, const Time &)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = Info();
  _info.size = info.size;
  pixel::TYPE type = pixel::type(info.pixel);
  switch (type)
  {
    case pixel::U10:
    case pixel::F16: 
    case pixel::F32: type = pixel::U16; break;
    default: break;
  }
  _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
  _info.endian = mem::MSB;
  DJV_DBG_PRN("info = " << _info);

  _image.set(_info);
}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.

  _io.open(in, djv::io::WRITE);
  
  // Header.
  
  const int channels = _image.channels();
  const int bytes = pixel::channel_bytes(_image.pixel());

  uint16_t magic = 474;
  _io.set_u16(magic);
  
  _io.set_u8(COMPRESS_RLE == _options.compress);
  _io.set_u8(bytes);
  
  const int dimension = 1 == channels ? (1 == _image.h() ? 1 : 2) : 3;
  _io.set_u16(dimension);
  
  _io.set_u16(_image.w());
  _io.set_u16(_image.h());
  _io.set_u16(channels);
  _io.set_u32(0);
  _io.set_u32(1 == bytes ? 255 : 65535);
  _io.position(512);
  
  // Initialize RLE tables.

  if (_options.compress)
  {
    const int table_size = _image.h() * channels;
    DJV_DBG_PRN("rle table size = " << table_size);
    _rle_offset.size(table_size);
    _rle_size.size(table_size);
    _io.seek(table_size * 2 * 4);
  }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("compress = " << _options.compress);

  // Open.
  
  _open(_file.get(frame));

  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    p = &_image;
  }
  
  _tmp.set(p->info());
  data::planar_deinterleave(*p, &_tmp);
  
  if (! _options.compress)
  {
    _io.set(_tmp.data(), _tmp.data_bytes());
  }
  else
  {
    const int w = _tmp.w(), h = _tmp.h();
    const int channels = pixel::channels(_tmp.pixel());
    const int bytes = pixel::channel_bytes(_tmp.pixel());
    mem::Buffer<uint8_t> scanline(w * bytes * 2);

    for (int c = 0; c < channels; ++c)
      for (int y = 0; y < h; ++y)
      {
        const size_t size = rle_pack(
          _tmp.data() + (c * h + y) * w * bytes, scanline(), w, bytes);
        _rle_offset()[y + c * h] = uint32_t(_io.position());
        _rle_size()[y + c * h] = uint32_t(size);
        _io.set(scanline.get(), size);
      }

    _io.position(512);
    _io.set_u32(_rle_offset(), h * channels);
    _io.set_u32(_rle_size(), h * channels);
  }
  
  _io.close();
}

}
}
}
}

