/*------------------------------------------------------------------------------
 lib/djv_image/io_tga_save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_tga_save.cxx"

#include "io_tga.h"

namespace djv {
namespace image {
namespace io {
namespace tga
{

Save::Save()
{
  _io.endian(mem::endian() != mem::LSB);
}

Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* tga compress (value) - Compression: %1. Default: %2.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_compress(), ", ").
    arg(str::serialize(_options.compress, label_compress()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  DJV_DBG_OBJ2(_DBG, "Save::option");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("data = " << *data);
  
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, label_compress());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  DJV_DBG_OBJ2(_DBG, "Save::option");
  DJV_DBG_PRN("in = " << in);
  
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, label_compress());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &, const Time &)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = Info();
  _info.size = info.size;
  _info.pixel = pixel::pixel(pixel::format(info.pixel), pixel::U8);
  switch (pixel::format(_info.pixel))
  {
    case pixel::RGB:
    case pixel::RGBA: _info.color = image::data::BGR; break;
    default: break;
  }
  _info.endian = mem::LSB;
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.

  _io.open(in, djv::io::WRITE);
  
  // Header.
  
  uint8_t image_type = 0;
  switch (_image.channels())
  {
    case 1:
    case 2: image_type = 3; break;
    case 3:
    case 4: image_type = 2; break;
  }
  if (COMPRESS_RLE == _options.compress) image_type += 8;
  
  int alpha_bits = 0;
  switch (pixel::format(_image.info().pixel))
  {
    case pixel::LA:
    case pixel::RGBA: alpha_bits = 8; break;
    default: break;
  }
  const int orient = 0;
  const uint8_t descriptor = (alpha_bits & 15) | ((orient & 3) << 4);

  _io.set_u8(0);
  _io.set_u8(0);
  _io.set_u8(image_type);
  _io.set_u16(0);
  _io.set_u16(0);
  _io.set_u8(0);
  _io.set_u16(0);
  _io.set_u16(0);
  _io.set_u16(_image.w());
  _io.set_u16(_image.h());
  _io.set_u8(_image.channels() * 8);
  _io.set_u8(descriptor);
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _open(_file.get(frame));
  
  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    p = &_image;
  }
  
  if (! _options.compress)
  {
    _io.set(p->data(), p->data_bytes());
  }
  else
  {
    const int w = p->w(), h = p->h();
    const int channels = pixel::channels(p->info().pixel);
    mem::Buffer<uint8_t> scanline(w * channels * 2);

    for (int y = 0; y < h; ++y)
    {
      size_t size = rle_pack(p->data(0, y), scanline(), w, channels);
      _io.set(scanline.get(), size);
    }
  }

  _io.close();
}

}
}
}
}

