/*------------------------------------------------------------------------------
 lib/djv_ui/columns.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/columns.cxx"

#include "columns.h"
#include <djv_ui/style.h>
#include <djv_ui/icon.h>
#include <FL/Fl.H>
#include <FL/fl_draw.H>

namespace djv {
namespace ui {
namespace columns
{

Widget::Widget() :
  _sort(-1), _sort_reverse(false),
  _push(-1), _drag(-1)
{
  //text_font(style::global()->font_bold());
}

void Widget::name_list(const List<Str> & in)
{
  if (in == name_list()) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::name_list");
  DJV_DBG_PRN("in = " << in);

  _column_list.resize(in.size());
  for (uint i = 0; i < _column_list.size(); ++i)
  {
    _column_list[i].name = in[i];
    _column_list[i].column = 0;
    _column_list[i].size_hint = 0;
  }
  
  columns_update();
  redraw();

  sig.emit(size_list());
}

void Widget::size_list(const List<int> & in)
{
  if (in == size_list()) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::size_list");
  DJV_DBG_PRN("in = " << in);

  _column_list.resize(in.size());
  
  const List<int> list = this->size_list();
  int x = 0;
  for (uint i = 0; i < in.size(); ++i)
  {
    x += in[i] ? in[i] : list[i];
    _column_list[i].column = x - 1;
  }
  
  columns_update();
  redraw();

  sig.emit(size_list());
}

List<Str> Widget::name_list() const
{
  List<Str> out;
  out.resize(_column_list.size());
  for (uint i = 0; i < _column_list.size(); ++i)
    out[i] = _column_list[i].name;
  return out;
}

List<int> Widget::size_list() const
{
  List<int> out;
  out.resize(_column_list.size());
  int x = 0;
  for (uint i = 0; i < _column_list.size(); ++i) {
    out[i] = _column_list[i].column - x + 1;
    x += out[i];
  }
  return out;
}

void Widget::sort(int in)
{
  if (in == _sort) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::sort");
  DJV_DBG_PRN("in = " << in);

  _sort = in;
  redraw();
}

void Widget::sort_reverse(bool in)
{
  if (in == _sort_reverse) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::sort_reverse");
  DJV_DBG_PRN("in = " << in);

  _sort_reverse = in;
  redraw();
}

int Widget::handle(int in)
{
  const Box2i & geom = this->geom();
  const Vec2i mouse(Fl::event_x(), Fl::event_y());
  switch (in)
  {
    case FL_PUSH:
    {
      DJV_DBG_OBJ2(_DBG, "Widget::handle");
      _drag = handle_drag(mouse.x - geom.x, _column_list);
      if (-1 == _drag) _push = handle_push(mouse.x - geom.x, _column_list);
      DJV_DBG_PRN("push = " << _push);
      DJV_DBG_PRN("drag = " << _drag);
      redraw();
    } return 1;
  
    case FL_RELEASE:
    {
      const int emit = intersect(geom, mouse) ? _push : -1;
      _push = _drag = -1;
      redraw();
      if (emit != -1) action_sig.emit(emit);
    } return 1;
    
    case FL_DRAG:
    {
      if (-1 == _drag) break;
      
      DJV_DBG_OBJ2(_DBG, "Widget::handle");
      DJV_DBG_PRN("drag = " << _drag);
      DJV_DBG_PRN("columns = " << size_list());
      
      const size_t size = _column_list.size();
      
      int min = 0, max = geom.w - 1;
      for (int i = 0; i <= _drag; ++i)
        min += _column_list[i].size_hint;
      for (uint i = _drag + 1; i < size; ++i)
        max -= _column_list[i].size_hint;
      
      _column_list[_drag].column = math::clamp(mouse.x - geom.x, min, max);

      // Push columns.
      
      for (int i = _drag - 1; i >= 0; --i)
        _column_list[i].column = math::min(
          _column_list[i].column,
          _column_list[i + 1].column - _column_list[i + 1].size_hint);
      for (uint i = _drag + 1; i < size; ++i)
        _column_list[i].column = math::max(
          _column_list[i].column,
          _column_list[i - 1].column + _column_list[i].size_hint);

      columns_update();
      redraw();
      
      sig.emit(size_list());
      
    } return 1;
    
    case FL_MOVE:
    {
      if (handle_drag(mouse.x - geom.x, _column_list) != -1)
        ui::cursor(FL_CURSOR_WE);
      else
        ui::cursor(FL_CURSOR_DEFAULT);
    } return 1;
    
    case FL_LEAVE:
    {
      ui::cursor(FL_CURSOR_DEFAULT);
    } return 1;
  }
  return widget::Widget::handle(in);
}

int Widget::handle_push(int in, const List<Column> & list)
{
  for (int i = 0; i < int(list.size()); ++i)
    if (in < list[i].column) return i;
  return -1;
}

int Widget::handle_drag(int in, const List<Column> & list)
{
  for (int i = 0; i < int(list.size()) - 1; ++i)
    if (math::intersect(in, list[i].column - style::global()->size_pick(),
      list[i].column + style::global()->size_pick()))
      return i;
  return -1;
}

void Widget::draw()
{
  DJV_DBG_OBJ2(_DBG, "Widget::draw");
  DJV_DBG_PRN("sort = " << _sort);
  DJV_DBG_PRN("sort reverse = " << _sort_reverse);
  
  const Box2i & geom = this->geom();
  const List<int> size_list = this->size_list();
  DJV_DBG_PRN("columns = " << size_list);

  // Background.
  
  fl_font(text_font(), style::global()->font_size());
  fl_color(FL_BACKGROUND_COLOR);
  fl_rectf(geom.x, geom.y, geom.w, geom.h);
  
  const int margin = style::global()->margin_widget();
  int x = geom.x;
  for (uint i = 0; i < size_list.size(); ++i)
  {
    fl_draw_box(
      int(i) == _push ? FL_THIN_DOWN_BOX : FL_THIN_UP_BOX,
      x, geom.y, size_list[i], geom.h,
      color());

    // Label.
    
    fl_color(FL_FOREGROUND_COLOR);
    fl_draw(_column_list[i].name.c_str(), x + margin,
      geom.y + geom.h / 2 - fl_height() / 2 + fl_height() - fl_descent());
    
    // Sort icon.
    
    if (int(i) == _sort)
    {
      Fl_Image * icon = ui::icon::get(_sort_reverse ? "tool_dec" : "tool_inc");
      icon->draw(x + size_list[i] - style::global()->size_tool_icon() - margin,
        geom.y + (geom.h - icon->h()) / 2);
    }
    
    x += size_list[i];
  }
}

void Widget::dirty()
{
  widget::Widget::dirty();

  fl_font(text_font(), style::global()->font_size());
  Vec2i size;
  for (uint i = 0; i < _column_list.size(); ++i)
  {
    const Vec2i label_size = str_size(_column_list[i].name);
    
    _column_list[i].size_hint =
      label_size.x +
      style::global()->size_tool_icon() +
      style::global()->margin_widget() * 3;

    size.x += _column_list[i].size_hint;
    size.y = math::max(label_size.y, size.y);
  }
  size_hint(Vec2i(size.x, size.y + style::global()->margin_widget() * 2));
}

void Widget::geom(const Box2i & in)
{
  bool dirty = in.size != geom().size;
  widget::Widget::geom(in);
  if (! dirty) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::geom");
  DJV_DBG_PRN("width = " << in.w);
  DJV_DBG_PRN("columns = " << size_list());
  
  // Resize first column.
  
  if (_column_list.size())
  {
    const int total = _column_list[_column_list.size() - 1].column;
    for (uint i = 0; i < _column_list.size(); ++i)
      _column_list[i].column += in.w - total - 1;
  }
  
  columns_update();

  sig.emit(size_list());
}

void Widget::columns_update()
{
  DJV_DBG_OBJ2(_DBG, "Widget::columns_update");
  DJV_DBG_PRN("columns = " << size_list());
  DJV_DBG_PRN("width = " << w());
  
  const Box2i & geom = this->geom();
  
  if (geom.w)
  {
    const size_t size = _column_list.size();
    List<int> size_list = this->size_list();
    
    // Check column minimum size.

    for (uint i = 0; i < size; ++i)
      size_list[i] = math::max(size_list[i], _column_list[i].size_hint);
    DJV_DBG_PRN("columns = " << size_list);
    
    // Fit columns within widget.
    
    int total = 0;
    for (uint i = 0; i < size; ++i)
      total += size_list[i];
    DJV_DBG_PRN("total = " << total);
    for (uint i = 0; i < size && total > geom.w; ++i)
    {
      int tmp = size_list[i];
      size_list[i] = math::max(size_list[i] - (total - geom.w),
        _column_list[i].size_hint);
      total -= tmp - size_list[i];
    }
    DJV_DBG_PRN("total = " << total);
    DJV_DBG_PRN("columns = " << size_list);
    
    int x = 0;
    for (uint i = 0; i < size; ++i)
    {
      x += size_list[i];
      _column_list[i].column = x - 1;
    }
    DJV_DBG_PRN("total = " << x);
  }
  
  DJV_DBG_PRN("columns = " << size_list());
}

}
}
}
