/*------------------------------------------------------------------------------
 plugin/djv_exr/load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_exr/load.cxx"

#include "plugin.h"
#include <djv_base/box.h>
#include <djv_image/time.h>
#include <ImfChannelList.h>
#include <ImfStandardAttributes.h>
#include <ImfHeader.h>
#include <ImfInputFile.h>
#include <ImfRgbaYca.h>
#include <algorithm>

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_load() { return new Load; }
}

Load::Load() : _f(0) {}
Load::~Load() { _close(); }

plugin::Plugin * Load::copy() const
{ Load * out = new Load; out->_options = _options; return out; }

namespace
{
const Str _help =
"* exr color (value) - Input color space: %1. Default: %2.\n"
"* exr exposure (value) (defog) (knee low) (knee high) (gamma) - Exposure "
"values. Default: %3.\n";
}

Str Load::help() const
{
  return str::Format(_help).
    arg(label_color(), ", ").arg(str::serialize(_options.color, label_color())).
    arg(serialize(_options.exposure));
}

bool Load::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_COLOR]))
    return str::serialize(data, &_options.color, label_color());
  else if (str::compare(in, option_list()[OPTION_EXPOSURE]))
    return serialize(data, &_options.exposure);
  return false;
}

List<Str> Load::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COLOR]))
    return str::serialize(_options.color, label_color());
  else if (str::compare(in, option_list()[OPTION_EXPOSURE]))
    return serialize(_options.exposure);
  return List<Str>();
}

const List<Str> & Load::option_list() const { return label_load_option(); }

void Load::open(const File & in, List<image::Info> * info, image::Tag * tag,
  image::Time * time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start), info, tag);
  
  if (file::SEQ == _file.type)
  {
    time->start = _file.seq.start;
    time->end = _file.seq.end;
  }
}

namespace
{
typedef bool (Compare)(const Channel &, const Channel &);

inline bool compare(const Channel & a, const Channel & b)
{
  if (a.name.size() && b.name.size())
  {
    const char c = a.name[int(a.name.size()) - 1];
    if ('Y' == c) return true;
    if ('R' == c || 'G' == c || 'B' == c)
      return c > b.name[int(b.name.size()) - 1];
  }
  return a.name < b.name;
}

image::pixel::FORMAT _format(List<Channel> * in)
{
  image::pixel::FORMAT out = image::pixel::FORMAT(0);
  
  uint size = 1;
  if (in->size())
    for (; size < in->size(); ++size)
      if ((*in)[size].type != (*in)[size - 1].type) break;

  switch (size)
  {
    case 1: out = image::pixel::L; break;
    case 2: out = image::pixel::LA; break;
    case 3: out = image::pixel::RGB; break;
    default: out = image::pixel::RGBA; break;
  }
  
  std::sort(in->begin(), in->end(), compare);
  return out;
}
}

void Load::_open(const File & in, List<image::Info> * info, image::Tag * tag)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);
  
try
{  
  _close();
  _f = new Imf::InputFile(in.get().c_str());
  
  // Information.
  
  const Imf::ChannelList & imf_channel_list = _f->header().channels();
  std::set<std::string> imf_layer_list;
  imf_channel_list.layers(imf_layer_list);

  image::Info _info;

  _info.size = imf_box(_f->header().dataWindow()).size;
  _info.mirror_y = true;

  List<Channel> channel_list;
  for (Imf::ChannelList::ConstIterator i = imf_channel_list.begin();
    i != imf_channel_list.end(); ++i)
  {
    const Str name = i.name();
    const Str::size_type index = name.find('.');
    if (index != Str::npos)
      if (index != 0 || index != int(name.size()) - 1) continue;
    channel_list += Channel(name, imf_type(i.channel().type),
      Vec2i(i.channel().xSampling, i.channel().ySampling));
  }
  
  image::pixel::FORMAT format = _format(&channel_list);
  image::pixel::TYPE type = image::pixel::TYPE(0);
  List<Channel> _channel_list;
  for (int i = 0; i < image::pixel::channels(format); ++i)
  {
    const Channel channel = channel_list.pop_front();
    type = math::max(channel.type, type);
    _channel_list += channel;
  }
  if (! image::pixel::pixel(format, type, &_info.pixel))
    throw Error(name(), image::io::error_unsupported);
  DJV_DBG_PRN("pixel = " << _info.pixel);

  info->resize(1 + channel_list.size() + imf_layer_list.size());
  (*info)[0] = _info;
  DJV_DBG_PRN("layers = " << info->size());
  
  _layer_list.clear();
  _layer_list += _channel_list;

  // XXX
    
  if (_channel_list.size() >= 3)
    if ("Y" == _channel_list[0].name && "RY" == _channel_list[1].name &&
      "BY" == _channel_list[2].name)
      (*info)[0].color = image::data::YCRCB;

  // Image tags.
  
  const List<Str> & label_tag = ::label_tag();
  const List<Str> & image_label_tag = image::tag::label();
  
  if (Imf::hasOwner(_f->header()))
    (*tag)[image_label_tag [image::tag::CREATOR]] =
      Imf::ownerAttribute(_f->header()).value();
  if (Imf::hasComments(_f->header()))
    (*tag)[image_label_tag [image::tag::DESCRIPTION]] =
      Imf::commentsAttribute(_f->header()).value();
  if (Imf::hasCapDate(_f->header()))
    (*tag)[image_label_tag [image::tag::TIME]] =
      Imf::capDateAttribute(_f->header()).value();
  if (Imf::hasutcOffset(_f->header()))
    (*tag)[image_label_tag [image::tag::UTC_OFFSET]] =
      str::float_to_str(Imf::utcOffsetAttribute(_f->header()).value());
  
  if (Imf::hasLongitude(_f->header()))
    (*tag)[label_tag[TAG_LONGITUDE]] =
      str::float_to_str(Imf::longitudeAttribute(_f->header()).value());
  if (Imf::hasLatitude(_f->header()))
    (*tag)[label_tag[TAG_LATITUDE]] =
      str::float_to_str(Imf::latitudeAttribute(_f->header()).value());
  if (Imf::hasAltitude(_f->header()))
    (*tag)[label_tag[TAG_ALTITUDE]] =
      str::float_to_str(Imf::altitudeAttribute(_f->header()).value());
  
  if (Imf::hasFocus(_f->header()))
    (*tag)[label_tag[TAG_FOCUS]] =
      str::float_to_str(Imf::focusAttribute(_f->header()).value());
  if (Imf::hasExpTime(_f->header()))
    (*tag)[label_tag[TAG_EXPOSURE]] =
      str::float_to_str(Imf::expTimeAttribute(_f->header()).value());
  if (Imf::hasAperture(_f->header()))
    (*tag)[label_tag[TAG_APERTURE]] =
      str::float_to_str(Imf::apertureAttribute(_f->header()).value());
  if (Imf::hasIsoSpeed(_f->header()))
    (*tag)[label_tag[TAG_ISO_SPEED]] =
      str::float_to_str(Imf::isoSpeedAttribute(_f->header()).value());

  if (Imf::hasChromaticities(_f->header())) {
    const Imf::Chromaticities data =
      Imf::chromaticitiesAttribute(_f->header()).value();
    (*tag)[label_tag[TAG_CHROMATICITIES]] = str::join(List<Str>() <<
      str::float_to_str(data.red.x) << str::float_to_str(data.red.y) <<
      str::float_to_str(data.green.x) << str::float_to_str(data.green.y) <<
      str::float_to_str(data.blue.x) << str::float_to_str(data.blue.y) <<
      str::float_to_str(data.white.x) << str::float_to_str(data.white.y), " ");
  }
  if (Imf::hasWhiteLuminance(_f->header()))
    (*tag)[label_tag[TAG_WHITE_LUMINANCE]] =
      str::float_to_str(Imf::whiteLuminanceAttribute(_f->header()).value());
  if (Imf::hasXDensity(_f->header()))
    (*tag)[label_tag[TAG_X_DENSITY]] =
      str::float_to_str(Imf::xDensityAttribute(_f->header()).value());
  
  if (Imf::hasKeyCode(_f->header())) {
    const Imf::KeyCode data = Imf::keyCodeAttribute(_f->header()).value();
    (*tag)[image_label_tag [image::tag::KEYCODE]] =
      image::time::keycode_to_str(
        data.filmMfcCode(), data.filmType(), data.prefix(), data.count(),
        data.perfOffset());
  }
  if (Imf::hasTimeCode(_f->header()))
    (*tag)[image_label_tag [image::tag::TIMECODE]] =
      image::time::timecode_to_str(
        Imf::timeCodeAttribute(_f->header()).value().timeAndFlags());
  
  // Layers.
  
  int layer = 0;
  for (; layer < channel_list.size(); ++layer)
  {
    (*info)[1 + layer] = _info;
    
    format = image::pixel::L;
    type = channel_list[layer].type;
    if (! image::pixel::pixel(format, type, &(*info)[1 + layer].pixel))
      throw Error(name(), image::io::error_unsupported);

    (*info)[1 + layer].name = channel_list[layer].name;
    
    _layer_list += List<Channel>() << channel_list[layer];
  }
  
  for (std::set<std::string>::const_iterator i = imf_layer_list.begin();
    i != imf_layer_list.end(); ++i, ++layer)
  {
    channel_list.clear();
    Imf::ChannelList::ConstIterator f, l;
    _f->header().channels().channelsInLayer(*i, f, l);
    for (Imf::ChannelList::ConstIterator k = f; k != l; ++k)
      channel_list += Channel(k.name(), imf_type(k.channel().type),
        Vec2i(k.channel().xSampling, k.channel().ySampling));

    (*info)[1 + layer] = _info;
    
    format = _format(&channel_list);
    type = image::pixel::TYPE(0);
    for (int k = 0; k < image::pixel::channels(format); ++k)
      type = math::max(channel_list[k].type, type);
    if (! image::pixel::pixel(format, type, &(*info)[1 + layer].pixel))
      throw Error(name(), image::io::error_unsupported);

    (*info)[1 + layer].name =
      Str(channel_list[0].name, 0, channel_list[0].name.rfind('.'));

    _layer_list += channel_list;
  }
  
  DJV_DBG_PRN("info = " << (*info)[0]);
  for (uint i = 1; i < info->size(); ++i)
    DJV_DBG_PRN("layer = " << (*info)[i]);

}
catch (const std::exception & error)
{
  throw Error(name(), error.what());
}
}

void Load::_close()
{
  if (_f) { delete _f; _f = 0; }
}

const image::Image * Load::get(int frame, int layer, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("layer = " << layer);
  DJV_DBG_PRN("proxy = " << proxy);

try
{  
  // Setup.
  
  if (layer < 0 || layer >= _layer_list.size())
    throw Error(name(), djv::io::error_read);
  
  const File file = File(_file.get(frame != -1 ? frame : _file.seq.start));
  DJV_DBG_PRN("file = " << file);

  if (COLOR_EXPOSURE == _options.color)
  {
    DJV_DBG_PRN("profile");
    _image.profile.type = image::profile::EXPOSURE;
    _image.profile.exposure = _options.exposure;
  }
  else
  {
    _image.profile = image::Profile();
  }
  
  /*Imf::Chromaticities chroma;
  Imath::V3f yw = Imf::RgbaYca::computeYw(chroma);
  DJV_DBG_PRN("yw = " << yw.x << " " << yw.y << " " << yw.z);*/

  const Cache cache(file, frame, proxy, layer);
  if (cache == _cache)
  {
    DJV_DBG_PRN("cached");
    return &_image;
  }
  _cache = cache;
  
  // Open.

  List<image::Info> info;
  _open(file, &info, &_image.tag);

  // Read.

  image::Data * p = proxy ? &_tmp : &_image;
  p->set(info[layer]);

  const bool flip = Imf::DECREASING_Y == _f->header().lineOrder();
  const Vec2i offset(_f->header().dataWindow().min.x,
    _f->header().dataWindow().min.y);
  DJV_DBG_PRN("flip = " << flip);
  DJV_DBG_PRN("data window = " << offset);
  
  const int channels = image::pixel::channels(info[layer].pixel);
  const int bytes = image::pixel::channel_bytes(info[layer].pixel);
  Imf::FrameBuffer frame_buffer;
  for (int c = 0; c < channels; ++c)
  {
    DJV_DBG_PRN("channel = " << _layer_list[layer][c].name);
    frame_buffer.insert(_layer_list[layer][c].name.c_str(),
      Imf::Slice(
        imf_type(image::pixel::type(p->pixel())),
        (char *)p->data() -
          (offset.y * info[layer].size.x * channels * bytes) -
          (offset.x * channels * bytes) +
          c * bytes,
        channels * bytes, info[layer].size.x * channels * bytes,
        _layer_list[layer][c].sampling.x, _layer_list[layer][c].sampling.y,
        0.0));
  }
  _f->setFrameBuffer(frame_buffer);

  if (flip)
    for (int y = 0; y < info[layer].size.y; ++y)
      _f->readPixels(offset.y + (info[layer].size.y - 1 - y));
  else
    _f->readPixels(offset.y, offset.y + info[layer].size.y - 1);

  if (proxy)
  {
    info[layer].size = image::data::proxy_scale(info[layer].size, proxy);
    info[layer].proxy = proxy;
    _image.set(info[layer]);
    image::data::proxy_scale(_tmp, &_image, proxy);
  }

}
catch (const std::exception & error)
{
  throw Error(name(), error.what());
}
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

