/*------------------------------------------------------------------------------
 plugin/djv_exr/plugin.h

 Industrial Light & Magic OpenEXR Image I/O
 ------------------------------------------
 
 Requires:
 * OpenEXR - http:://www.openexr.com

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_EXR_PLUGIN_H
#define DJV_EXR_PLUGIN_H

#include <djv_image/io.h>
#include <ImfPixelType.h>
#include <ImathBox.h>

namespace Imf
{
class InputFile;
class OutputFile;
}

using namespace djv;

//------------------------------------------------------------------------------
// Plugin Information

static const Str name = "EXR";

static const List<Str> extension_list = List<Str>() << ".exr";

//------------------------------------------------------------------------------
// Plugin Options

enum COLOR { COLOR_RAW, COLOR_EXPOSURE };
enum COMPRESS { COMPRESS_NONE, COMPRESS_PIZ, COMPRESS_ZIP, COMPRESS_RLE };

//------------------------------------------------------------------------------
// Image Tags

enum TAG { TAG_LONGITUDE, TAG_LATITUDE, TAG_ALTITUDE, TAG_FOCUS, TAG_EXPOSURE,
  TAG_APERTURE, TAG_ISO_SPEED, TAG_CHROMATICITIES, TAG_WHITE_LUMINANCE,
  TAG_X_DENSITY };

//------------------------------------------------------------------------------
// EXR Data

struct Channel
{
  Channel() : sampling(1) {}
  Channel(const Str & name, image::pixel::TYPE type, const Vec2i & sampling) :
    name(name), type(type), sampling(sampling) {}

  Str name;
  image::pixel::TYPE type;
  Vec2i sampling;
};

//------------------------------------------------------------------------------
// Load Plugin

class Load : public image::io::Load
{
public:

  Load();
  ~Load();
  
  plugin::Plugin * copy() const;

  Str name() const { return ::name; }
  const List<Str> & extension_list() const { return ::extension_list; }
  Str help() const;
  
  struct Options
  {
    Options() : color(COLOR_EXPOSURE) {}
    COLOR color;
    image::profile::Exposure exposure;
  };
  enum OPTION { OPTION_COLOR, OPTION_EXPOSURE };
  bool option(const Str &, List<Str> *);
  List<Str> option(const Str &) const;
  const List<Str> & option_list() const;
  
  void open(const File & in, image::Info * info, image::Tag * tag,
    image::Time * time) throw (Error)
    { image::io::Load::open(in, info, tag, time); }
  void open(const File &, List<image::Info> *, image::Tag *, image::Time *)
    throw (Error);
  const image::Image * get(int frame, int layer, int proxy) throw (Error);

private:
  
  void _open(const File &, List<image::Info> *, image::Tag *) throw (Error);
  void _close();
  
  Options _options;
  File _file;
  Imf::InputFile * _f;
  List<List<Channel> > _layer_list;
  image::Image _image;
  image::Data _tmp;
  Cache _cache;
};

//------------------------------------------------------------------------------
// Save Plugin

class Save : public image::io::Save
{
public:

  Save();
  ~Save();
  
  plugin::Plugin * copy() const;
  
  Str name() const { return ::name; }
  const List<Str> & extension_list() const { return ::extension_list; }
  Str help() const;
  
  struct Options
  {
    Options() : compress(COMPRESS_NONE) {}
    COMPRESS compress;
  };
  enum OPTION { OPTION_COMPRESS };
  bool option(const Str &, List<Str> *);
  List<Str> option(const Str &) const;
  const List<Str> & option_list() const;
  
  void open(const File &, const image::Info &, const image::Tag &,
    const image::Time &) throw (Error);
  void set(const image::Image &, int frame) throw (Error);

private:

  void _open(const File &, const image::Tag &) throw (Error);
  void _close();

  Options _options;
  File _file;
  Imf::OutputFile * _f;
  List<Str> _channel_list;
  image::Info _info;
  image::Data _tmp;
};

//------------------------------------------------------------------------------
// Conversion

Box2i imf_box(const Imath::Box2i &);

Imf::PixelType imf_type(image::pixel::TYPE);
image::pixel::TYPE imf_type(Imf::PixelType);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_color();
const List<Str> & label_compress();
const List<Str> & label_tag();
const List<Str> & label_load_option();
const List<Str> & label_save_option();

#endif

