/*------------------------------------------------------------------------------
 plugin/djv_exr/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_exr/save.cxx"

#include "plugin.h"
#include <djv_image/time.h>
#include <ImfChannelList.h>
#include <ImfCompressionAttribute.h>
#include <ImfStandardAttributes.h>
#include <ImfHeader.h>
#include <ImfOutputFile.h>

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() : _f(0) {}
Save::~Save() { _close(); }

plugin::Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* exr compress (value) - Compression: %1. Default: %2.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_compress(), ", ").
    arg(str::serialize(_options.compress, label_compress()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("data = " << *data);
  
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, label_compress());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, label_compress());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const image::Info & info, const image::Tag &,
  const image::Time &) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = image::Info();
  _info.size = info.size;
  _info.mirror_y = true;
  image::pixel::FORMAT format = image::pixel::format(info.pixel);
  image::pixel::TYPE type = image::pixel::F16;
  switch (type)
  {
    case image::pixel::F32: type = image::pixel::F32; break;
    default: break;
  }
  _info.pixel = image::pixel::pixel(format, type);
  DJV_DBG_PRN("info = " << _info);
  
  _tmp.set(_info);
}

void Save::_open(const File & in, const image::Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

try
{
  _close();
  
  // Header.
  
  Imf::Header header(_info.size.x, _info.size.y);
  
  switch (image::pixel::channels(_info.pixel))
  {
    case 1: _channel_list = List<Str>() << "Y"; break;
    case 2: _channel_list = List<Str>() << "Y" << "A"; break;
    case 3: _channel_list = List<Str>() << "R" << "G" << "B"; break;
    case 4: _channel_list = List<Str>() << "R" << "G" << "B" << "A"; break;
  }
  for (uint i = 0; i < _channel_list.size(); ++i)
    header.channels().insert(_channel_list[i].c_str(),
      image::pixel::type(_info.pixel) == image::pixel::F16 ?
      Imf::HALF : Imf::FLOAT);
  
  Imf::CompressionAttribute compress;
  switch (_options.compress)
  {
    case COMPRESS_NONE: compress = Imf::NO_COMPRESSION; break;
    case COMPRESS_PIZ: compress = Imf::PIZ_COMPRESSION; break;
    case COMPRESS_ZIP: compress = Imf::ZIP_COMPRESSION; break;
    case COMPRESS_RLE: compress = Imf::RLE_COMPRESSION; break;
  }
  header.insert(Imf::CompressionAttribute::staticTypeName(), compress);
  
  // Image tags.
  
  const List<Str> & label_tag = ::label_tag();
  const List<Str> & image_label_tag = image::tag::label();
  
  Str _tag = tag[image_label_tag[image::tag::CREATOR]];
  if (_tag.size()) addOwner(header, _tag);
  _tag = tag[image_label_tag[image::tag::DESCRIPTION]];
  if (_tag.size()) addComments(header, _tag);
  _tag = tag[image_label_tag[image::tag::TIME]];
  if (_tag.size()) addCapDate(header, _tag);
  _tag = tag[image_label_tag[image::tag::UTC_OFFSET]];
  if (_tag.size()) addutcOffset(header, str::str_to_float(_tag));
    
  _tag = tag[label_tag[TAG_LONGITUDE]];
  if (_tag.size()) addLongitude(header, str::str_to_float(_tag));
  _tag = tag[label_tag[TAG_LATITUDE]];
  if (_tag.size()) addLatitude(header, str::str_to_float(_tag));
  _tag = tag[label_tag[TAG_ALTITUDE]];
  if (_tag.size()) addAltitude(header, str::str_to_float(_tag));

  _tag = tag[label_tag[TAG_FOCUS]];
  if (_tag.size()) addFocus(header, str::str_to_float(_tag));
  _tag = tag[label_tag[TAG_EXPOSURE]];
  if (_tag.size()) addExpTime(header, str::str_to_float(_tag));
  _tag = tag[label_tag[TAG_APERTURE]];
  if (_tag.size()) addAperture(header, str::str_to_float(_tag));
  _tag = tag[label_tag[TAG_ISO_SPEED]];
  if (_tag.size()) addIsoSpeed(header, str::str_to_float(_tag));

  _tag = tag[label_tag[TAG_CHROMATICITIES]];
  if (_tag.size())
  {
    List<Str> tmp = str::split(_tag, " ");
    if (8 == tmp.size())
      addChromaticities(header, Imf::Chromaticities(
        Imath::V2f(str::str_to_float(tmp[0]), str::str_to_float(tmp[1])),
        Imath::V2f(str::str_to_float(tmp[2]), str::str_to_float(tmp[3])),
        Imath::V2f(str::str_to_float(tmp[4]), str::str_to_float(tmp[5])),
        Imath::V2f(str::str_to_float(tmp[6]), str::str_to_float(tmp[7]))));
  }
  _tag = tag[label_tag[TAG_WHITE_LUMINANCE]];
  if (_tag.size()) addWhiteLuminance(header, str::str_to_float(_tag));
  _tag = tag[label_tag[TAG_X_DENSITY]];
  if (_tag.size()) addXDensity(header, str::str_to_float(_tag));

  _tag = tag[image_label_tag[image::tag::KEYCODE]];
  if (_tag.size())
  {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    image::time::str_to_keycode(_tag, &id, &type, &prefix, &count, &offset);
    addKeyCode(header, Imf::KeyCode(id, type, prefix, count, offset));
  }
  _tag = tag[image_label_tag[image::tag::TIMECODE]];
  if (_tag.size()) addTimeCode(header, image::time::str_to_timecode(_tag));

  // Open.

  _f = new Imf::OutputFile(in.get().c_str(), header);

}
catch (const std::exception & error)
{
  throw Error(name(), error.what());
}
}

void Save::_close()
{
  if (_f) { delete _f; _f = 0; }
}

void Save::set(const image::Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  
try
{

  // Open.
  
  _open(_file.get(frame), in.tag);
  
  // Write.
  
  const image::Data * p = &in;
  if (p->info() != _info)
  {
    DJV_DBG_PRN("convert = " << _tmp);
    _tmp.zero();
    image::gl::data::copy(in, &_tmp);
    p = &_tmp;
  }

  const int w = p->w(), h = p->h();
  const int channels = p->channels();
  const int bytes = image::pixel::channel_bytes(p->pixel());

  Imf::FrameBuffer frame_buffer;
  for (int c = 0; c < channels; ++c)
    frame_buffer.insert(_channel_list[c].c_str(),
      Imf::Slice(
        imf_type(image::pixel::type(p->pixel())),
        (char *)p->data() + c * bytes,
        channels * bytes, w * channels * bytes, 1, 1,
        0.0));
  _f->setFrameBuffer(frame_buffer);

  _f->writePixels(h);

}
catch (const std::exception & err)
{
  throw Error(name(), err.what());
}
  _close();
}

