/*------------------------------------------------------------------------------
 plugin/djv_quicktime_unix/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_quicktime_unix/save.cxx"

#include "plugin.h"

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() : _f(0), _frame_duration(0) {}

plugin::Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* quicktime codec (value) - Codec: %1. Default: %2.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_codec(), ", ").arg(_options.codec);
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_CODEC]))
    return str::serialize(data, &_options.codec);
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_CODEC]))
    return str::serialize(_options.codec);
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const image::Info & info,
  const image::Tag & tag, const image::Time & time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("info = " << info);
  
  // Open.
  
  _file = in;
  _f = lqt_open_write(in.get().c_str(), LQT_FILE_QT);
  if (! _f)
    throw Error(name(), image::io::error_open);
  
  // Information.
  
  struct Codec_Info
  {
    Codec_Info(const Str & name) : p(0)
    { p = lqt_find_video_codec_by_name(name.c_str()); }
    
    ~Codec_Info()
    { if (p) lqt_destroy_codec_info(p); }
    
    lqt_codec_info_t ** p;
  
  } codec_info(_options.codec);
  
  if (! codec_info.p)
    throw Error(name(), error_codec);
  
  _frame_duration = time.speed.duration;
  
  // XXX Return value?

  lqt_set_video(_f, 1, info.size.x, info.size.y, _frame_duration,
    time.speed.scale, codec_info.p[0]);
  
  _info = image::Info();
  _info.size = info.size;
  _info.pixel = info.pixel;
  static int cmodel_list [] = {
    BC_RGB888,
    BC_BGR888,
    BC_RGB161616,
    BC_RGBA8888,
    BC_RGBA16161616,
    LQT_COLORMODEL_NONE };
  const int cmodel = lqt_get_best_colormodel(_f, 0, cmodel_list);
  switch (cmodel)
  {
    case BC_BGR888: _info.color = image::data::BGR;
    case BC_RGB888: _info.pixel = image::pixel::RGB_U8; break;
    case BC_RGB161616: _info.pixel = image::pixel::RGB_U16; break;
    case BC_RGBA8888: _info.pixel = image::pixel::RGBA_U8; break;
    case BC_RGBA16161616: _info.pixel = image::pixel::RGBA_U16; break;
    default: throw Error(name(), error_colormodel);
  }
  lqt_set_cmodel(_f, 0, cmodel);
  DJV_DBG_PRN("info = " << _info);
  
  const List<Str> & label_tag = image::tag::label();
  Str _tag = tag[label_tag[image::tag::CREATOR]];
  if (_tag.size()) lqt_set_artist(_f, (char *)_tag.c_str());
  _tag = tag[label_tag[image::tag::DESCRIPTION]];
  if (_tag.size()) lqt_set_comment(_f, (char *)_tag.c_str());
  _tag = tag[label_tag[image::tag::COPYRIGHT]];
  if (_tag.size()) quicktime_set_copyright(_f, (char *)_tag.c_str());

  _image.set(_info);
}

void Save::close() throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::close");

  if (_f) { quicktime_close(_f); _f = 0; }
}

void Save::set(const image::Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);

  // Write.
  
  const image::Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    image::gl::data::copy(in, &_image);
    p = &_image;
  }

  struct Rows
  {
    Rows(const image::Data * in) : p(0)
    {
      p = new unsigned char * [in->h() + 1];
      for (int y = 0; y < in->h(); ++y)
        p[y] = const_cast<image::pixel::U8_T *>(in->data(0, in->h() - 1 - y));
    }
    
    ~Rows() { if (p) delete [] p; }
    
    unsigned char ** p;
    
  } rows(p);
  
  lqt_encode_video(_f, rows.p, 0, frame * _frame_duration);
}

